#!/usr/bin/env python3

"""
Construct/compile the computation graph, and optionally save it to some file.
There are various options/variations for what task and what conditions you can create the graph,
e.g. for training, forwarding, search, or also step-by-step execution over a recurrent layer.
"""

from __future__ import annotations

import typing
import os
import sys
import argparse
import contextlib
import tensorflow as tf
from tensorflow.python.framework import graph_io
from tensorflow.python.util import nest

import _setup_returnn_env  # noqa
import returnn.__main__ as rnn
from returnn.log import log
from returnn.config import Config
import returnn.util.basic as util
from returnn.util.basic import NotSpecified, Entity
import returnn.tf.compat as tf_compat
import returnn.tf.util.basic as tf_util
from returnn.tf.util.basic import Data, Dim, CollectionKeys
from returnn.tf.network import TFNetwork, ExternData
from returnn.tf.layers.basic import LayerBase, register_layer_class, SourceLayer
from returnn.tf.layers.base import WrappedInternalLayer

# noinspection PyProtectedMember
from returnn.tf.layers.rec import RecLayer, _SubnetworkRecCell, _TemplateLayer


config = None  # type: typing.Optional[Config]


def init(config_filename, log_verbosity, device):
    """
    :param str config_filename: filename to config-file
    :param int log_verbosity:
    :param str device:
    """
    rnn.init_better_exchook()
    print("Using config file %r." % config_filename)
    assert os.path.exists(config_filename)
    rnn.init_config(
        config_filename=config_filename,
        extra_updates={
            "use_tensorflow": True,
            "log": None,
            "log_verbosity": log_verbosity,
            "task": __file__,  # just extra info for the config
            "device": device,
        },
    )
    global config
    config = rnn.config
    rnn.init_log()
    print("Returnn compile-tf-graph starting up.", file=log.v1)
    rnn.returnn_greeting()
    rnn.init_backend_engine()
    assert util.BackendEngine.is_tensorflow_selected(), "this is only for TensorFlow"
    rnn.init_faulthandler()


def create_graph(train_flag, eval_flag, search_flag, net_dict):
    """
    :param bool train_flag:
    :param bool eval_flag:
    :param bool search_flag:
    :param dict[str,dict[str]] net_dict:
    :return: adds to the current graph, and then returns the network
    :rtype: returnn.tf.network.TFNetwork
    """
    print("Loading network, train flag %s, eval flag %s, search flag %s" % (train_flag, eval_flag, search_flag))
    from returnn.tf.engine import Engine
    from returnn.tf.network import TFNetwork

    network, updater = Engine.create_network(
        config=config,
        rnd_seed=1,
        train_flag=train_flag,
        eval_flag=eval_flag,
        search_flag=search_flag,
        net_dict=net_dict,
    )
    assert isinstance(network, TFNetwork)
    return network


@contextlib.contextmanager
def helper_variable_scope():
    """
    :return: separate scope from the current name scope, such that variables are not treated as model params
    :rtype: tf.VariableScope
    """
    with tf_util.reuse_name_scope("IO", absolute=True) as scope:
        yield scope


class SubnetworkRecCellSingleStep(_SubnetworkRecCell):
    """
    Adapts :class:`_SubnetworkRecCell` such that we execute only a single step.
    Used by :class:`RecStepByStepLayer`. See :class:`RecStepByStepLayer` for further documentation.
    """

    def __init__(self, **kwargs):
        self._parent_layers = {}  # type: typing.Dict[str,WrappedInternalLayer]
        self._parent_dim_tags = {}  # type: typing.Dict[Dim,Dim]
        self._parent_replace_deps = []  # type: typing.List[typing.Tuple[RecStepByStepLayer.StateVar,Data]]
        super(SubnetworkRecCellSingleStep, self).__init__(**kwargs)
        extern_data_copy = ExternData()
        extern_data_copy.data.update({k: v.copy_template() for (k, v) in self.net.extern_data.data.items()})
        self.net_delayed_update = TFNetwork(
            name="%s(delayed-update)" % self.net.name,
            extern_data=extern_data_copy,
            train_flag=self.net.train_flag,
            search_flag=self.net.search_flag,
            eval_flag=False,
            inside_rec_time_dim=self.time_dim_tag,
            control_flow_ctx=self.net.control_flow_ctx,
            absolute_name_prefix=self.net.get_absolute_name_prefix(),
            parent_net=self.parent_net,
        )
        self.net_delayed_update.is_root_in_ctx = True
        self.delayed_state_update_op = None  # type: typing.Optional[tf.Operation]
        self.state_update_op = None  # type: typing.Optional[tf.Operation]

    def _maybe_delay_tiled(self, state_var, output):
        """
        :param RecStepByStepLayer.StateVar state_var:
        :param Data output:
        :rtype: tf.Tensor
        """
        assert isinstance(state_var, RecStepByStepLayer.StateVar)
        assert isinstance(output, Data)
        rec_layer = self.parent_rec_layer
        assert isinstance(rec_layer, RecStepByStepLayer)
        if rec_layer.construction_state == rec_layer.ConstructionState.GetSources:
            self._parent_replace_deps.append((state_var, output))  # only in case the layer is accessed elsewhere later
            # No tiling yet because we cannot infer the in-loop batch dim. done in _construct.
            # Also use the state var read because we will use this later inside the loop.
            x = state_var.read()
        elif rec_layer.construction_state == rec_layer.ConstructionState.Init:
            self._parent_replace_deps.append((state_var, output))
            with tf.control_dependencies([state_var.init_op()]):
                x = tf.identity(
                    output.placeholder,
                    name="state_var_before_loop_depend_on_init__%s"
                    % tf_util.get_valid_scope_name_from_str(state_var.name),
                )
        elif rec_layer.construction_state == rec_layer.ConstructionState.InLoop:
            x = self._tiled(output, state_var.read())
        else:
            raise ValueError("unexpected construction state %r" % rec_layer.construction_state)
        dim_tag = Dim.get_tag_from_size_tensor(output.placeholder)
        if dim_tag:
            dim_tag.set_tag_on_size_tensor(x, same_as_before=True)
        output.placeholder = x
        return x

    def _tiled(self, output, x):
        """
        :param Data output:
        :param tf.Tensor x:
        :rtype: tf.Tensor
        """
        rec_layer = self.parent_rec_layer
        assert isinstance(rec_layer, RecStepByStepLayer)
        return tf_util.tile_transposed(x, axis=output.batch_dim_axis, multiples=rec_layer.get_parent_tile_multiples())

    def get_parent_dim_tag(self, dim_tag):
        """
        :param Dim dim_tag:
        :rtype: Dim
        """
        if dim_tag.dimension is not None:
            return dim_tag  # as-is, not dynamic
        if dim_tag.is_batch_dim():
            return dim_tag  # as-is
        if dim_tag in self._parent_dim_tags:
            return self._parent_dim_tags[dim_tag]
        rec_layer = self.parent_rec_layer
        assert isinstance(rec_layer, RecStepByStepLayer)
        state_var = rec_layer.create_state_var(
            name="base_size_%s" % tf_util.get_valid_scope_name_from_str(dim_tag.description),
            initial_value=dim_tag.dyn_size,
        )
        dim_tag_dyn_size_ext = dim_tag.dyn_size_ext.copy()
        with tf_util.same_control_flow_ctx(dim_tag_dyn_size_ext.placeholder), tf_util.reuse_name_scope(
            "", absolute=True
        ):
            # Need to create new tensor currently for the set_tag_on_size_tensor logic.
            # (set_tag_on_size_tensor is going to be removed at some point but currently we need it.)
            dim_tag_dyn_size_ext.placeholder = tf.identity(
                dim_tag_dyn_size_ext.placeholder, name=dim_tag_dyn_size_ext.placeholder.op.name + "_copy_new_dim_tag"
            )
        # Copy dim tag.
        dim_tag_ = Dim(
            kind=dim_tag.kind,
            description=dim_tag.description + "_base_state_var",
            dimension=None,
            dyn_size_ext=dim_tag_dyn_size_ext,
            auto_generated=dim_tag.auto_generated,  # keep same is_equal behavior
            batch=dim_tag.batch,
        )
        dim_tag_.set_tag_on_size_tensor(dim_tag_dyn_size_ext.placeholder)
        self._maybe_delay_tiled(state_var, dim_tag_.dyn_size_ext)
        self._parent_dim_tags[dim_tag] = dim_tag_
        return dim_tag_

    def _get_parent_layer(self, layer_name):
        """
        :param str layer_name: without "base:" prefix
        :rtype: WrappedInternalLayer
        """
        if layer_name in self._parent_layers:
            return self._parent_layers[layer_name]
        layer = super(SubnetworkRecCellSingleStep, self)._get_parent_layer(layer_name)
        rec_layer = self.parent_rec_layer
        if rec_layer is None:  # at template construction
            return layer
        assert isinstance(rec_layer, RecStepByStepLayer)
        output = layer.output.copy()
        new_dim_tags = []
        for i, dim_tag in enumerate(output.dim_tags):
            new_dim_tags.append(self.get_parent_dim_tag(dim_tag))
        output = output.copy_template_new_dim_tags(new_dim_tags=new_dim_tags, keep_special_axes=True)
        output.placeholder = layer.output.placeholder
        state_var = rec_layer.create_state_var(
            name="base_value_%s" % layer_name, initial_value=output.placeholder, data_shape=output
        )
        self._maybe_delay_tiled(state_var, output)
        layer = WrappedInternalLayer(
            name=layer_name, network=self.parent_net, output=output, base_layer=layer, debug_type_name="rec-single-step"
        )
        self._parent_layers[layer_name] = layer
        return layer

    def get_sources(self, sources):
        """
        :param list[str] sources:
        :rtype: list[WrappedInternalLayer]
        """
        return [self._get_parent_layer(layer_name) for layer_name in sources]

    def _set_construction_state_in_loop(self):
        rec_layer = self.parent_rec_layer
        assert isinstance(rec_layer, RecStepByStepLayer)
        rec_layer.set_construction_state_in_loop()

        for args in self._parent_replace_deps:
            self._maybe_delay_tiled(*args)

    def _while_loop(self, cond, body, loop_vars, shape_invariants):
        """
        :param function cond:
        :param function body:
        :param T loop_vars:
        :param S shape_invariants:
        :rtype: T

        def body(i, net_vars, acc_tas, seq_len_info=None)
          tf.Tensor i: loop counter, scalar
          net_vars: the accumulator values. see also self.get_init_loop_vars()
          list[tf.TensorArray] acc_tas: the output accumulator TensorArray
          (tf.Tensor,tf.Tensor)|None seq_len_info: tuple (end_flag, seq_len)
          return: [i + 1, a_flat, tas]: the updated counter + new accumulator values + updated TensorArrays
            rtype (tf.Tensor, object, list[tf.TensorArray])

        def cond(i, net_vars, acc_tas, seq_len_info=None)
          ...
          return: tf.Tensor bool, True or False
        """
        cell = self
        rec_layer = self.parent_rec_layer
        assert isinstance(rec_layer, RecStepByStepLayer)
        self.net_delayed_update.parent_layer = rec_layer

        if len(loop_vars) == 3:
            i, net_vars, acc_tas = loop_vars
            seq_len_info = None
        else:
            i, net_vars, acc_tas, seq_len_info = loop_vars
            seq_len_info_ = rec_layer.create_state_vars_recursive(("end_flag", "dyn_seq_len"), seq_len_info)
            seq_len_info = nest.map_structure(lambda state_var: state_var.read(), seq_len_info_)
        initial_i = i
        i = rec_layer.create_state_var("i", initial_i).read()

        # Go through layers with state, check their (direct + indirect) dependencies,
        # whether they depend on any choice vars.
        layers_with_state = set(self._initial_outputs.keys()).union(self._initial_extra_outputs.keys())
        layers_cur_iteration = set()
        layers_delayed = set()
        layer_deps_by_layer = {}  # type: typing.Dict[str, typing.Set[str]]
        layers_delayed_prev_deps = set()
        choice_layers = set()
        for layer_name in layers_with_state:
            template_layer = self.layer_data_templates[layer_name]
            queue = [template_layer]
            visited = set()
            prev_frame_deps = set()
            choice_deps = set()
            source_deps = set()
            while queue:
                cur = queue.pop(0)
                if cur in visited:
                    continue
                visited.add(cur)
                if cur.layer_class_type is ChoiceStateVarLayer:
                    choice_deps.add(cur.name)
                    continue
                if cur.layer_class_type is SourceLayer:
                    source_deps.add(cur.name)
                    continue
                for dep in cur.cur_frame_dependencies:
                    if dep in visited:
                        continue
                    if self.net not in dep.network.get_network_hierarchy():
                        continue  # ignore layers from parent network
                    assert isinstance(dep, _TemplateLayer)
                    queue.append(dep)
                prev_frame_deps.update([dep.name for dep in cur.prev_frame_dependencies])
            layer_deps_by_layer[layer_name] = prev_frame_deps
            if not choice_deps:
                layers_cur_iteration.add(layer_name)
            else:
                print(
                    "Delayed: Layer %r depends on choices %r, deps on prev frame %r"
                    % (layer_name, choice_deps, prev_frame_deps),
                    file=log.v4,
                )
                if source_deps:
                    raise NotImplementedError(
                        "Delayed layers with source dependencies (%s) not supported yet" % source_deps
                    )
                layers_delayed.add(layer_name)
                layers_delayed_prev_deps.update(prev_frame_deps)
                choice_layers.update(choice_deps)
        for choice_layer in choice_layers:
            assert choice_layer not in layers_cur_iteration
            layers_delayed.add(choice_layer)  # if not yet added, add now because we anyway need them
            rec_layer.add_stochastic_var(choice_layer)

        # See _SubnetworkRecCell.get_output().body().
        # net_vars is (prev_outputs_flat, prev_extra_flat).
        # prev_outputs_flat corresponds to the dict self._initial_outputs, specifically sorted(self._initial_outputs)
        # where the keys are the layer names.
        # prev_extra is always expected to be batch-major.
        # prev_extra_flat corresponds to the dict self._initial_extra_outputs,
        # specifically sorted(self._initial_extra_outputs),
        # where the keys are the layer names.
        init_outputs_flat, init_extra_flat = net_vars
        assert len(init_outputs_flat) == len(self._initial_outputs)
        assert len(init_extra_flat) == len(self._initial_extra_outputs)
        init_outputs = {k: v for k, v in zip(sorted(self._initial_outputs.keys()), init_outputs_flat)}
        init_extra = {
            k: util.dict_zip(sorted(self._initial_extra_outputs[k]), v)
            for (k, v) in zip(sorted(self._initial_extra_outputs), init_extra_flat)
        }

        # noinspection PyShadowingNames
        class _LayerStateHelper:
            def __init__(self, layer_name, prefix):
                """
                :param str layer_name:
                :param str prefix:
                """
                self.layer_name = layer_name
                self.prefix = prefix

                initial_values = {}
                data_shapes = {}
                if layer_name in init_outputs:
                    initial_values["output"] = init_outputs[layer_name]
                    data_shapes["output"] = cell.layer_data_templates[layer_name].output
                if layer_name in init_extra:
                    initial_values["extra"] = init_extra[layer_name]
                    data_shapes["extra"] = nest.map_structure(
                        lambda s: None, init_extra[layer_name]
                    )  # should be batch-major
                assert isinstance(rec_layer, RecStepByStepLayer)
                self.state_vars = rec_layer.create_state_vars_recursive(
                    name_prefix="%s/%s" % (prefix, layer_name), initial_values=initial_values, data_shapes=data_shapes
                )
                self._reads_once = None

            def reset_reads_once(self):
                """
                Next get_reads_once will create new op.
                """
                self._reads_once = None

            def get_reads_once(self):
                """
                :return: same structure as state vars with the actual reads, type tf.Tensor.
                    makes sure they are created once
                """
                if self._reads_once is not None:
                    return self._reads_once
                self._reads_once = self.reads()
                return self._reads_once

            def reads(self):
                """
                :return: same structure as state vars with the actual reads, type tf.Tensor.
                """
                return nest.map_structure(lambda state_var: state_var.read(), self.state_vars)

            def assigns_flat(self, state):
                """
                :param state: same structure as state vars
                :return: list of tf.Assign ops
                :rtype: list[tf.Operation]
                """

                def _map(state_var, state_value):
                    assert isinstance(state_var, RecStepByStepLayer.StateVar)
                    assert isinstance(state_value, tf.Tensor)
                    return state_var.assign(state_value)

                assert isinstance(rec_layer, RecStepByStepLayer)
                nest.assert_same_structure(state, self.state_vars)
                return nest.flatten(nest.map_structure(_map, self.state_vars, state))

            @staticmethod
            def get_from_net(net, layer_name, prev=False):
                """
                :param TFNetwork net:
                :param str layer_name:
                :param bool prev:
                :return: same structure as state vars, type tf.Tensor
                """
                layer = net.layers[("prev:" + layer_name) if prev else layer_name]
                out = {}
                if layer_name in init_outputs:
                    out["output"] = layer.output.placeholder
                if layer_name in init_extra:
                    out["extra"] = layer.rec_vars_outputs
                return out

        with tf.name_scope("state_delayed"):
            layers_prev_prev = {}
            for layer_name in layers_delayed.union(layers_delayed_prev_deps):
                layers_prev_prev[layer_name] = _LayerStateHelper(layer_name, "state_delayed")

        with tf.name_scope("state"):
            layers_prev = {}
            for layer_name in layers_cur_iteration:
                layers_prev[layer_name] = _LayerStateHelper(layer_name, "state")

        # We are ignoring acc_tas (the tensor arrays).

        self._set_construction_state_in_loop()

        # Create read ops now to maybe use them in the conditional branch.
        for layer_name in layers_cur_iteration:
            layers_prev[layer_name].get_reads_once()
        for layer_name in layers_delayed.union(layers_delayed_prev_deps):
            layers_prev_prev[layer_name].get_reads_once()

        with tf.name_scope("delayed_state_update"):
            # noinspection PyShadowingNames
            def _delayed_state_update():
                self._construct_custom(
                    net=self.net_delayed_update,
                    prev_state={
                        layer_name: layers_prev_prev[layer_name].get_reads_once()
                        for layer_name in layers_delayed.union(layers_delayed_prev_deps)
                    },
                    cur_state={
                        layer_name: layers_prev[layer_name].get_reads_once() for layer_name in layers_cur_iteration
                    },
                    needed_outputs=layers_delayed,
                )

                # Make sure all state vars dependencies are read first before we reassign them.
                control_deps = []
                for layer_name in layers_delayed.union(layers_delayed_prev_deps):
                    control_deps += nest.flatten(layers_prev_prev[layer_name].get_reads_once())
                for layer_name in choice_layers:
                    control_deps.append(self.net_delayed_update.layers[layer_name].output.placeholder)

                # Now reassign the delayed state vars.
                with tf.control_dependencies(control_deps):
                    ops = []
                    # Update all delayed state vars, i.e. prev:prev:layer to prev:layer.
                    for layer_name in layers_delayed.union(layers_delayed_prev_deps):
                        if layer_name in layers_cur_iteration:
                            # Can simply copy it. "prev:layer" is already in it.
                            state = layers_prev[layer_name].get_reads_once()
                        else:
                            assert layer_name in layers_delayed
                            state = _LayerStateHelper.get_from_net(self.net_delayed_update, layer_name)
                        ops += layers_prev_prev[layer_name].assigns_flat(state)

                return tf.group(*ops)

            self.delayed_state_update_op = tf.cond(tf.greater(i, initial_i), _delayed_state_update, tf.no_op)

            # We should not use control dependencies on delayed_state_update_op
            # because delayed_state_update_op will be update_ops and therefore executed first.

            # Now we should update net_vars accordingly

            for v in layers_prev_prev.values():
                v.reset_reads_once()
                v.get_reads_once()

            outputs_flat = []
            for layer_name in sorted(self._initial_outputs):
                if layer_name in layers_cur_iteration:
                    outputs_flat.append(layers_prev[layer_name].get_reads_once()["output"])
                else:
                    assert layer_name in layers_delayed
                    outputs_flat.append(layers_prev_prev[layer_name].get_reads_once()["output"])  # This is updated now.
            extra_flat = []
            for layer_name, v in sorted(self._initial_extra_outputs.items()):
                if layer_name in layers_cur_iteration:
                    state = layers_prev[layer_name].get_reads_once()["extra"]
                else:
                    assert layer_name in layers_delayed
                    state = layers_prev_prev[layer_name].get_reads_once()["extra"]  # This is updated now.
                assert isinstance(v, dict) and isinstance(state, dict)
                assert set(state.keys()) == set(v.keys())
                extra_flat.append(util.sorted_values_from_dict(state))
            net_vars = (outputs_flat, extra_flat)

        state_update_ops = []
        with tf.name_scope("cond"):
            s = rec_layer.create_state_var("cond", tf.constant(True))
            state_update_ops.append(s.assign(cond(i, net_vars, acc_tas, seq_len_info, allow_inf_max_len=True)))

        with tf.name_scope("body"):
            # The body function is locally defined in _SubnetworkRecCell.get_output().
            # This function then calls _SubnetworkRecCell._construct() which will do the subnet construction.
            res = body(i, net_vars, acc_tas, seq_len_info)
            assert len(res) == len(loop_vars)
            if len(res) == 3:
                i, net_vars, acc_tas = res
                seq_len_info = None
            else:
                i, net_vars, acc_tas, seq_len_info = res
        if seq_len_info:
            state_update_ops += rec_layer.assign_state_vars_recursive_flatten(("end_flag", "dyn_seq_len"), seq_len_info)
        if rec_layer.rec_step_by_step_opts["update_i_in_graph"]:
            state_update_ops.append(rec_layer.state_vars["i"].assign(i))

        # Assign new state.
        for layer_name in layers_cur_iteration:
            state = _LayerStateHelper.get_from_net(self.net, layer_name)
            state_update_ops += layers_prev[layer_name].assigns_flat(state)
        self.state_update_op = tf.group(*state_update_ops)

        return res

    def _construct_custom(self, net, prev_state, cur_state, needed_outputs):
        """
        This is a simplified version of _SubnetworkRecCell._construct without search logic.

        :param TFNetwork net:
        :param prev_state:
        :param cur_state:
        :param needed_outputs:
        """
        assert isinstance(net, TFNetwork)

        prev_layers = {}  # type: typing.Dict[str,_TemplateLayer]

        # noinspection PyShadowingNames
        def _add_predefined_layer(layer_name, state_dict, prev):
            assert isinstance(state_dict, dict)
            assert set(state_dict.keys()).issubset({"output", "extra"})
            try:
                layer = self.layer_data_templates[name].copy_as_prev_time_frame(
                    prev_output=state_dict.get("output", None), rec_vars_prev_outputs=state_dict.get("extra", None)
                )
            except Exception as exc:  # Some sanity check might have failed or so.
                self._handle_construct_exception(description="in-loop init of prev layer %r" % name, exception=exc)
                raise
            layer.network = net
            if prev:
                prev_layers[layer_name] = layer
            net.layers[("prev:%s" % layer_name) if prev else layer_name] = layer

        for name, state_dict in prev_state.items():
            _add_predefined_layer(name, state_dict, prev=True)
        for name, state_dict in cur_state.items():
            _add_predefined_layer(name, state_dict, prev=False)

        # noinspection PyShadowingNames
        def get_layer(name):
            """
            :param str name: layer name
            :rtype: LayerBase
            """
            if name.startswith("prev:"):
                return prev_layers[name[len("prev:") :]]
            if name.startswith("base:"):
                return self._get_parent_layer(name[len("base:") :])
            # noinspection PyBroadException
            try:
                layer = net.construct_layer(self.net_dict, name=name, get_layer=get_layer)
                if name == "end":
                    # Special logic for the end layer, to always logical_or to the prev:end layer.
                    assert layer.output.shape == (), "%s: 'end' layer %r unexpected shape" % (
                        self.parent_rec_layer,
                        layer,
                    )
                    prev_end_layer = net.layers["prev:end"]
                    choices = layer.get_search_choices()
                    if choices:
                        prev_end_layer = choices.translate_to_this_search_beam(prev_end_layer)
                    with tf.name_scope("end_flag"):
                        layer.output.placeholder = tf.logical_or(
                            prev_end_layer.output.placeholder, layer.output.placeholder
                        )
                return layer
            except Exception as exc:
                self._handle_construct_exception(description="in-loop construction of layer %r" % name, exception=exc)
                raise

        # Go through needed_outputs, e.g. "output".
        # And prev_layers_needed because they might not be resolved otherwise.
        for layer_name in sorted(needed_outputs):
            get_layer(layer_name)

    def _construct(self, prev_outputs, prev_extra, i, data=None, inputs_moved_out_tas=None, needed_outputs=("output",)):
        """
        This is called from within the body of the while loop
        (`tf.while_loop` in :class:`RecLayer` but here only a single step)
        to construct the subnetwork, which is performed step by step.

        :param dict[str,tf.Tensor] prev_outputs:
        :param dict[str,dict[str,tf.Tensor]] prev_extra:
        :param tf.Tensor i: loop counter. scalar, int32, current step (time)
        :param dict[str,tf.Tensor] data: via tensor arrays
        :param dict[str,tf.TensorArray]|None inputs_moved_out_tas:
        :param set[str] needed_outputs:
        """
        # The TensorArrays depend on the base network (base state vars)
        # which would not have tiling enabled at the time they were constructed.
        # So we add tiling here.
        assert data is not None
        for key, value in list(data.items()):
            if key != "source":
                continue  # ignore other sources currently...
            assert key in self.net.extern_data.data
            data_ = self.net.extern_data.data[key]
            data[key] = self._tiled(data_, value)

        # Fixup all base state var dim tags in the template layers.
        # These templates are used to define prev layers,
        # e.g. like prev accum att weights, so we might need the parent dyn sizes now.
        all_dyn_parent_dim_tags = set()
        for layer in self.parent_net.get_root_network().get_all_layers_deep():
            if self.net in layer.network.get_network_hierarchy():
                continue
            all_dyn_parent_dim_tags.update(
                [dim for dim in layer.output.dim_tags if dim.dimension is None and not dim.is_batch_dim()]
            )
        for layer in self.layer_data_templates.values():
            assert isinstance(layer, _TemplateLayer)
            dim_tags = list(layer.output.dim_tags)
            for _i, tag in enumerate(dim_tags):
                if tag in all_dyn_parent_dim_tags:
                    dim_tags[_i] = self.get_parent_dim_tag(tag)
            layer.output = layer.output.copy_template_new_dim_tags(dim_tags)
            layer.kwargs["output"] = layer.output

        super(SubnetworkRecCellSingleStep, self)._construct(
            prev_outputs=prev_outputs,
            prev_extra=prev_extra,
            i=i,
            data=data,
            inputs_moved_out_tas=inputs_moved_out_tas,
            needed_outputs=needed_outputs,
        )


class RecStepByStepLayer(RecLayer):
    """
    Represents a single step of :class:`RecLayer`.
    The purpose is to execute a single step only.
    This also takes care of all needed state, and stochastic (maybe latent) variables (via :class:`ChoiceLayer`).
    All the state is kept in *state variables*, such that you can avoid feeding/fetching.
    Instead, any decoder implementation using this must explicitly assign the state variables.
    Stochastic variables (:class:`ChoiceLayer`) are breakpoints where an external application
    can implement custom logic to select hypotheses.
    So this can be used to implement beam search, or to implement a custom decoder in an external application.
    RASR (Sprint) is one such example which makes use of this for decoding.

    The necessary meta information about the list of state vars,
    ops for specific calculation steps, etc.
    is stored in TF graph collections and also in a JSON file.
    RASR currently only uses the TF graph collections.

    There are different kinds of state vars:
    - Base state vars (encoder or so), depends only on the input data, not updated by the decoder loop.
    - decoder_input_vars: Stochastic choice state vars, updated by the decoder loop.
    - decoder_output_vars: Stochastic scores state vars, updated by the decoder loop.
    - (Deterministic) loop state vars, updated by the decoder loop.

    RASR does the following logic (currently only a single stochastic var supported)::

      # initial state vars
      encode_ops(input_placeholders=...)  # -> assign (base and loop) state vars

      # state_vars are only the loop state vars here, not the base state vars.
      # Store initial loop state values.
      # Note that this is not really necessary always (including the logic described here)
      # because we anyway then assign them again in the first iteration of the decoder loop
      # and never need them otherwise again.
      state_vars.readout()

      # Main decoder loop
      while seq_not_ended(...):

        for hyp in current_hyps:  # (in practice, this is partially batched)

          state_vars.assign(...)  # for current hyp (might be skipped in first decoder loop iteration)
          decoder_input_vars.assign(...)  # for current hyp, the previous choice

          update_ops(decoder_input_vars)  # -> assign/update potentially some loop state vars

          # Assume only a single decode_ops here (single ChoiceLayer).
          # decoder_input_vars contains prev choices
          decode_ops(state_vars, decoder_input_vars)  # -> assign decoder_output_vars (scores)

          # note that decoder_input_vars for current hyp is still the previous (!) choice

          post_update_ops(state_vars, decoder_input_vars)  # -> assign new state_vars

          state_vars.readout()  # readout and store them for this hyp

        < select new hyps and prune some away, based on scores >

    Because RASR is the main application, we adopt the recurrence logic of the RecLayer to be compatible to RASR.

    Note that the update_ops, decode_ops and post_update_ops all depend on the previous state vars (obviously)
    or maybe decode_ops would depend on some updated state vars, updated by update_ops,
    or maybe post_update_ops would depend on some updated state vars, updated by decode_ops or update_ops.
    The logic flow is as outlined above.
    And all ops depends on the previous (!) choice vars and not on the current choice vars.
    This is in contrast to RETURNN, where the final loop state vars are potentially based on the current choice var.

    One way to solve this is by delaying the state vars by one iteration.
    So when "prev state vars" would normally refer to "prev:layer" (also including hidden state),
    now it would refer to "prev:prev:layer".
    Then for the decode_ops, we need to make sure first that we transfer "prev:prev:layer" to "prev:layer"
    and then we can calculate it as normal.
    Even when we merge the post_update_ops with the decode_ops, in many models,
    this still requires to do a lot (most) computation redundantly twice
    because for computing the decode_ops, we do most of the same computation which we would do for computing
    "layer" based on "prev:layer", but "layer" is not used here because we delayed the state vars by one iteration.

    We might keep all state vars twice, once for the current iteration (when possible),
    and once for the delayed iteration (when needed).
    But this is bad because this would cause a lot of overhead when the decoder needs to copy around the state vars.
    This is the case for RASR which handles high number of hypotheses, where the state vars are stored in CPU memory.

    As mentioned before, decode_ops shares a lot of the computation with updating "layer" based on "prev:layer".
    Or said differently, the calculation of "layer" is often not dependent on the current choice var.
    When we can figure out the dependencies exactly, we can know which state vars can stay in the current iteration,
    and which need to be delayed.
    There are potentially also some state vars which we need to keep both for the current and previous iteration
    in order to be able to compute the delayed update when this depends on it.

    So, for the loop state vars, we have two cases:

    - It refers to the current iteration.
    - It refers to the previous iteration.

    For some layer (which is part of the state because it is accessed via "prev:layer" or has hidden state),
    we can have three cases:

    - Loop state vars referring to the current iteration.
    - Loop state vars referring to the previous iteration.
    - Loop state vars for both the current and previous iteration.

    The update of delayed state vars must take extra care in the first frame.
    It is initialized with the initial state (layers initial_output, and initial hidden state),
    and in the first frame, it would just not update this.
    Only from the second frame on, it is delayed.

    How to we implement this logic here?

    First we aggressively search for layers which can stay in the current iteration,
    i.e. which do not depend on any of the :class:`ChoiceLayer`s.
    The update calculation from "prev:layer" to "layer" would be jointly with decode_ops
    to avoid redundant computation.
    This should be the first decode_ops if there are multiple.
    This makes the post_update_ops obsolete.

    For delayed state vars, before we do anything else,
    in the beginning of the iteration, we need to update "prev:prev:layer" to "prev:layer",
    or also set "prev:layer" based on the prev choices (e.g. "prev:output").
    For this, we need to know which dependencies on the state are needed for "layer" ("prev:layer"),
    and then all such dependencies are needed to be stored as delayed state vars as well.
    Then everything else in the iteration can be done as usual.
    This first step will be the update_ops.

    So first we need to figure out which state vars are needed of what kind.
    Then we must construct the update_ops for one part of layers.
    Then we must construct the decode_ops (including merged post_update_ops) for another set of layers,
    and there can be overlap both in the inputs and outputs,
    so these must be two separate constructions.

    Note that RASR only needs to know about the loop state vars, not the base state vars,
    and the stochastic state vars are also handled separately.
    RASR will re-assign and readout the loop state vars in every loop iteration.

    ---

    If you want to implement beam search in an external application which uses the compiled graph,
    you would compile the graph with search_flag disabled (such that RETURNN does not do any related logic),
    enable this recurrent step-by-step compilation, and then do the following TF session runs:

    * Call the initializers of all the state variables, which also includes everything from the base (e.g. encoder),
      while feeding in the placeholders for the input features.
      This is "init_op" in the info json, "encode_ops" in the TF graph collections.
      All further session runs should not need any feed values. All needed state should be in state vars.
      This will init all state vars, except stochastic_var_*.
    * For each decoder step:
      * For each stochastic variable (both latent variables and observable variables),
        in the order as in "stochastic_var_order" in the info json:
        - This calculation depends on all state vars
          (except of stochastic_var_scores_* and only the dependent other stochastic_var_choice_*).
          I.e. all those state vars must be assigned accordingly.
        - Calculate `"stochastic_var_scores_%s" % name` (which are the probabilities in +log space).
          This is "calc_scores_op" in the info json, "decode_ops" in the TF graph collections.
        - Do a choice, build up a beam of hypotheses.
        - Set the stochastic state var `"stochastic_var_choice_%s" % name` to the selected values (label indices).
        - If the beam has multiple items, i.e. the batch dimension changed, you must make sure
          that all further used state variables will also have the same batch dim.
      * Do a single session run for the next values of these state vars:
        "i", "end_flag" (if existing), "dyn_seq_len" (if existing), "state_*" (multiple vars).
        These are also the state vars which will get updated in every further recurrent step.
        The "base_*" state vars are always kept (although you might need to update the batch dim),
        and the "stochastic_var_*" state vars have the special logic for the stochastic variables.
        This is "next_step_op" in the info json.

    Thus, the info json should contain the following:

    * "state_vars": dict[str, dict[str, Any]]. name -> dict with:
      * "var_op": str
      * "shape": tuple[int|None,...]. including batch dim. always batch-major, or scalar.
      * "dtype": str
      The state vars usually are:
      * "i"
      * "end_flag" (if existing)
      * "dyn_seq_len" (if existing)
      * "state_*" (multiple vars)
      * "base_*" (multiple vars). only once via init_op, maybe then batch-tiled, but otherwise not changed for once seq.
      * "stochastic_var_*" (multiple vars)
      However, no specific assumptions should be made on any of these.
      It should not matter to the decoder what we have here.
    * "stochastic_vars": dict[str,dict[str, str]]. name -> dict with:
      * "calc_scores_op": str. op to calculate the scores state var
      * "scores_state_var": str. == "stochastic_var_scores_%s" % name
      * "choice_state_var": str. == "stochastic_var_choice_%s" % name
    * "stochastic_var_order": list[str]. the order of the stochastic vars.

    * "init_op": str. op. the initializer for all state vars, including encoder.
    * "next_step_op: str. op. the update op for all state vars.

    ---

    This layer class derives from :class:`RecLayer` and adopts the logic to be able
    to construct the mentioned ops for step-by-step execution.

    The final ops are constructed in :func:`post_compile`
    and put into corresponding TF graph collections and the info json.

    However, the main construction logic happens before, in :func:`__init__`,
    and then further in :func:`SubnetworkRecCellSingleStep._while_loop`.

    ---

    See https://github.com/rwth-i6/returnn/pull/874 for some discussion.
    """

    layer_class = "rec_step_by_step"
    SubnetworkRecCell = SubnetworkRecCellSingleStep

    class ConstructionState:
        """construction states"""

        GetSources = Entity("get_sources")
        Init = Entity("init")
        InLoop = Entity("in_body")

    @classmethod
    def prepare_compile(cls, rec_layer_name, net_dict, opts):
        """
        :param str rec_layer_name:
        :param dict[str,dict[str]] net_dict:
        :param dict[str] opts:
        :return: nothing, will prepare globally, and modify net_dict in place
        """
        register_layer_class(RecStepByStepLayer)
        register_layer_class(ChoiceStateVarLayer)
        assert rec_layer_name in net_dict
        rec_layer_dict = net_dict[rec_layer_name]
        assert rec_layer_dict["class"] == "rec"
        assert isinstance(rec_layer_dict["unit"], dict)
        for key, layer_dict in list(rec_layer_dict["unit"].items()):
            assert isinstance(layer_dict, dict)
            if layer_dict["class"] == "choice":
                layer_dict["class"] = ChoiceStateVarLayer.layer_class
        rec_layer_dict["class"] = RecStepByStepLayer.layer_class
        rec_layer_dict["rec_step_by_step_opts"] = opts

    @classmethod
    def post_compile(cls, rec_layer_name, network, output_file_name=None):
        """
        :param str rec_layer_name:
        :param TFNetwork network:
        :param str|None output_file_name: via command line --rec_step_by_step_output_file
        """
        assert rec_layer_name in network.layers
        rec_layer = network.layers[rec_layer_name]
        assert isinstance(rec_layer, RecStepByStepLayer)
        cell = rec_layer.cell
        assert isinstance(cell, SubnetworkRecCellSingleStep)
        rec_sub_net = cell.net
        assert isinstance(rec_sub_net, TFNetwork)
        info = {"state_vars": {}, "stochastic_var_order": [], "stochastic_vars": {}}

        print("State vars:")
        for name, var in sorted(rec_layer.state_vars.items()):
            assert isinstance(name, str)
            assert isinstance(var, RecStepByStepLayer.StateVar)
            print(" %s: %r, shape %s, dtype %s" % (name, var.var.op.name, var.var.shape, var.var.dtype.base_dtype.name))
            info["state_vars"][name] = {
                "var_op": var.var.op.name,
                "shape": [int(d) if d is not None else None for d in var.var_data_shape.batch_shape],
                "dtype": var.var.dtype.base_dtype.name,
            }

        # global_vars were used by the computation graph in some earlier version.
        # We do not make use of global_vars anymore because this is redundant with the state vars.
        # We anyway create the collection here such that older RASR binaries still work fine.
        # See https://github.com/rwth-i6/returnn/pull/874.
        global_vars_coll = tf_compat.v1.get_collection_ref("global_vars")

        # Encoder and all decoder state initializers.
        encode_ops_coll = tf_compat.v1.get_collection_ref("encode_ops")
        # We must make sure that the non-base loop vars and all their dependencies,
        # are initialized after we have initialized the base loop vars.
        # A control dependency on the var.init_op is not enough as the init_op is only the tf.assign, not more.
        # We make sure in the get_parent_layer via the construction_state.
        init_ops = []
        for name, var in sorted(rec_layer.state_vars.items()):
            assert isinstance(name, str)
            assert isinstance(var, RecStepByStepLayer.StateVar)
            if not name.startswith("stochastic_var_"):
                init_ops.append(var.init_op())
        init_op = tf.group(*init_ops, name="rec_step_by_step_init_op")
        info["init_op"] = init_op.name
        encode_ops_coll.append(init_op)

        # Based on some decoder state and encoder, calculate the next scores for the next stochastic variable.
        decode_ops_coll = tf_compat.v1.get_collection_ref("decode_ops")  # calculates the scores
        decoder_output_vars_coll = tf_compat.v1.get_collection_ref("decoder_output_vars")  # scores
        decoder_input_vars_coll = tf_compat.v1.get_collection_ref("decoder_input_vars")  # choices
        print("Stochastic vars, and their order:")
        # additional flexibility to match RASR's specific order of decoder inputs feeding
        if rec_layer.reverse_stochastic_var_order:
            stochastic_var_order = rec_layer.stochastic_var_order[::-1]
        else:
            stochastic_var_order = rec_layer.stochastic_var_order
        for name in stochastic_var_order:
            print(" %s" % name)
            info["stochastic_var_order"].append(name)
            choice_layer = rec_sub_net.layers[name]
            assert isinstance(choice_layer, ChoiceStateVarLayer)
            # Usually all choices depend on the scores.
            # However, some legacy configs used prev choices which are fed as-is because they are known.
            # See comment in ChoiceStateVarLayer.
            if name == rec_layer.stochastic_var_order[-1]:
                assert choice_layer.score_dependent
            if choice_layer.score_dependent:
                calc_scores_op = rec_layer.state_vars["stochastic_var_scores_%s" % name].final_op()
                info["stochastic_vars"][name] = {
                    "calc_scores_op": calc_scores_op.name,
                    "scores_state_var": "stochastic_var_scores_%s" % name,
                    "choice_state_var": "stochastic_var_choice_%s" % name,
                }
                if name == rec_layer.stochastic_var_order[-1]:
                    # The last decode op can be merged with the state updates (post_update_ops).
                    calc_scores_op = tf.group(calc_scores_op, cell.state_update_op)
                decode_ops_coll.append(calc_scores_op)
                decoder_output_vars_coll.append(rec_layer.state_vars["stochastic_var_scores_%s" % name].var)
            else:
                info["stochastic_vars"][name] = {"choice_state_var": "stochastic_var_choice_%s" % name}
            decoder_input_vars_coll.append(rec_layer.state_vars["stochastic_var_choice_%s" % name].var)

        update_ops_coll = tf_compat.v1.get_collection_ref("update_ops")
        update_ops_coll.append(cell.delayed_state_update_op)
        # Based on the decoder state, encoder, and choices, calculate the next state.
        tf_compat.v1.get_collection_ref("post_update_ops")  # leave empty; merged with last decode op
        state_vars_coll = tf_compat.v1.get_collection_ref(CollectionKeys.STATE_VARS)
        for name, var in sorted(rec_layer.state_vars.items()):
            assert isinstance(name, str)
            assert isinstance(var, RecStepByStepLayer.StateVar)
            if not name.startswith("stochastic_var_") and not name.startswith("base_") and name != "cond":
                if var.orig_data_shape.have_batch_axis():
                    state_vars_coll.append(var.var)
                else:
                    global_vars_coll.append(var.var)

        import json

        info_str = json.dumps(info, sort_keys=True, indent=2)
        if not output_file_name:
            print("No rec-step-by-step output file name specified, not storing this info.")
            print("JSON:")
            print(info_str)
        else:
            with open(output_file_name, "w") as f:
                f.write(info_str)
            print("Stored rec-step-by-step info JSON in file:", output_file_name)

    class StateVar:
        """
        Represents a state variable, i.e. either a state, a choice, or encoder state, etc.
        """

        def __init__(self, parent, name, initial_value, data_shape):
            """
            :param RecStepByStepLayer parent:
            :param str name:
            :param tf.Tensor|None initial_value:
              initial_value might have dim 1 in variable dimensions (which are not the batch-dim-axis),
              see get_rec_initial_output, which should be fine for broadcasting.
            :param Data data_shape:
              Describes the shape of initial_value, and also what we store as self.orig_data_shape,
              and what we return by self.read().
              If it is not a scalar, and batch-dim-axis > 0, the created variable will still be in batch-major.
            """
            self.parent = parent
            self.name = name
            self.orig_data_shape = data_shape
            self.var_data_shape = (
                data_shape.copy_as_batch_major() if data_shape.batch_dim_axis is not None else data_shape
            )
            del data_shape
            self.orig_initial_value = initial_value
            if initial_value is not None and self.orig_data_shape.batch_dim_axis not in (0, None):
                x = self.orig_data_shape.copy()
                x.placeholder = initial_value
                x = x.copy_compatible_to(self.var_data_shape)
                initial_value = x.placeholder
            self.var_initial_value = initial_value
            del initial_value
            # Note: Don't use `initializer` of `tf.get_variable` directly, because
            # it uses _try_guard_against_uninitialized_dependencies internally,
            # which replace references to variables in `initial_value`
            # with references to the variable's initialized values.
            # This is not what we want.
            # Also, it has a cycle check which is extremely inefficient and basically just hangs.
            # Instead, some dummy initializer. The shape should not matter.
            zero_initializer = tf_util.zeros_dyn_shape(
                shape=self.var_data_shape.batch_shape, dtype=self.var_data_shape.dtype
            )
            with helper_variable_scope():
                self.var = tf_compat.v1.get_variable(
                    name=tf_util.get_valid_scope_name_from_str(name), initializer=zero_initializer, validate_shape=False
                )  # type: tf.Variable
            self.var.set_shape(self.var_data_shape.batch_shape)
            self._init_op = None
            print("New state var %r: %s, shape %s" % (name, self.var, self.var_data_shape))
            self.final_value = None  # type: typing.Optional[tf.Tensor]

        def __repr__(self):
            return "<StateVar %r, shape %r, initial %r>" % (self.name, self.var_data_shape, self.orig_initial_value)

        def set_final_value(self, final_value):
            """
            :param tf.Tensor final_value:
            """
            assert self.final_value is None
            assert isinstance(final_value, tf.Tensor)
            self.final_value = final_value

        def read(self):
            """
            :return: tensor in the format of self.orig_data_shape
            :rtype: tf.Tensor
            """
            value = self.var.read_value()
            if self.orig_data_shape.batch_dim_axis in (0, None):
                return value  # should be the right shape
            # Need to convert from self.var_data_shape to self.orig_data_shape (different batch-dim-axis).
            x = self.var_data_shape.copy()
            x.placeholder = value
            x = x.copy_compatible_to(self.orig_data_shape)
            return x.placeholder

        def init_op(self):
            """
            :return: op which assigns self.var_initial_value to self.var
            :rtype: tf.Operation
            """
            if self._init_op:
                return self._init_op
            assert self.var_initial_value is not None
            self._init_op = tf_compat.v1.assign(
                self.var, self.var_initial_value, name="init_state_var_%s" % self.name
            ).op
            return self._init_op

        def final_op(self):
            """
            :return: op which does self.var.assign(self.final_value) (final value maybe converted)
            :rtype: tf.Operation
            """
            assert self.final_value is not None
            return self.assign(self.final_value)

        def assign(self, value):
            """
            :return: op which does self.var.assign(value) (value maybe converted)
            :rtype: tf.Operation
            """
            from returnn.tf.util.basic import find_ops_path_output_to_input

            feed_tensors = []
            for data in self.parent.network.extern_data.data.values():
                feed_tensors.append(data.placeholder)
                feed_tensors.extend(data.size_placeholder.values())
            path = find_ops_path_output_to_input(fetches=value, tensors=feed_tensors)
            assert not path, "There should be no path from extern data to %s final op value, but there is:\n%s" % (
                self,
                "\n".join(map(repr, path)),
            )
            if self.orig_data_shape.batch_dim_axis not in (0, None):
                x = self.orig_data_shape.copy()
                x.placeholder = value
                x = x.copy_compatible_to(self.var_data_shape)
                value = x.placeholder
            return tf_compat.v1.assign(self.var, value, name="final_state_var_%s" % self.name).op

    def __init__(
        self,
        _orig_sources,
        sources,
        network,
        name,
        output,
        unit,
        axis,
        rec_step_by_step_opts,
        reverse_stochastic_var_order=False,
        **kwargs,
    ):
        """
        :param str|list[str]|None _orig_sources:
        :param list[LayerBase] sources:
        :param returnn.tf.network.TFNetwork network:
        :param str name:
        :param Data output:
        :param SubnetworkRecCellSingleStep unit:
        :param Dim axis:
        :param dict[str] rec_step_by_step_opts:
        :param bool reverse_stochastic_var_order:
        """
        assert isinstance(unit, SubnetworkRecCellSingleStep)
        kwargs = kwargs.copy()
        kwargs["optimize_move_layers_out"] = False
        self.state_vars = {}  # type: typing.Dict[str,RecStepByStepLayer.StateVar]
        self.stochastic_var_order = []  # type: typing.List[str]
        self.stochastic_vars = {}  # type: typing.Dict[str,RecStepByStepLayer.StochasticVar]
        self._parent_tile_multiples = None  # type: typing.Optional[tf.Tensor]
        self.construction_state = self.ConstructionState.GetSources
        self.reverse_stochastic_var_order = reverse_stochastic_var_order
        self.rec_step_by_step_opts = rec_step_by_step_opts

        # Some early assigns needed for set_parent_layer, and that __repr__ works.
        self.network = network
        self.name = name
        self.output = output
        unit.set_parent_layer(self)  # early assign, so that unit.get_parent_layer() works
        if _orig_sources:
            # Make sure we resolve sources again through our get_parent_layer logic.
            if isinstance(_orig_sources, (list, tuple)):
                assert len(sources) == len(_orig_sources)
            else:
                assert isinstance(_orig_sources, str)
                assert len(sources) == 1
                _orig_sources = [_orig_sources]
            assert unit.parent_rec_layer is self  # this is needed such that get_sources and get_parent_layer works
            # get_sources will call get_parent_layer which will register it properly as a base state var.
            # Note that we are not yet inside the loop, thus this will not enable tiling,
            # and thus the resulting TensorArray will not have tiling.
            # Thus, we will post-edit the subnet extern_data ("data:source") in our derived _construct to add tiling.
            sources = unit.get_sources(_orig_sources)
        else:
            assert not sources

        # axis might be an existing axis from the encoder (e.g. transducer),
        # so we also need to wrap it.
        assert isinstance(axis, Dim)
        if axis and sources and sources[0].base_layer.output.have_dim_tag(axis):
            base_axis = axis
            axis = unit.get_parent_dim_tag(base_axis)
            unit.time_dim_tag = axis
            # noinspection PyProtectedMember
            unit._time_dim_tags.add(axis)
            unit.net._inside_rec_time_dim = axis
            if output.have_dim_tag(base_axis):
                output = output.copy_template_replace_dim_tag(
                    axis=output.get_axis_from_description(base_axis), new_dim_tag=axis
                )

        self.construction_state = self.ConstructionState.Init
        if self.have_base_state_vars():
            self._set_global_batch_dim(self.get_batch_dim_from_base_state_var())
        # This will eventually get to SubnetworkRecCellSingleStep._while_loop.
        super(RecStepByStepLayer, self).__init__(
            sources=sources, network=network, name=name, output=output, unit=unit, axis=axis, **kwargs
        )

    def set_construction_state_in_loop(self):
        """
        Set that we entered the body.
        """
        self.construction_state = self.ConstructionState.InLoop

        # Some layers make explicit use of the (global data) batch-dim,
        # which they can get via TFNetwork.get_data_batch_dim().
        # This will add a dependency on the external data, which we want to avoid.
        # We can avoid that by taking the batch dim instead from one of the other states.
        # Note that this would be wrong in beam search.
        self._set_global_batch_dim(self.get_batch_dim_from_loop_state_var())

    def _set_global_batch_dim(self, batch_dim):
        """
        :param tf.Tensor batch_dim:
        """
        self.network.get_global_batch_info().dim = batch_dim

    def get_parent_tile_multiples(self):
        """
        :rtype: tf.Tensor
        """
        if self._parent_tile_multiples is not None:
            return self._parent_tile_multiples
        with tf_util.reuse_name_scope("parent_tile_multiples", absolute=True):
            base_batch_dim = self.get_batch_dim_from_base_state_var()
            loop_batch_dim = self.get_batch_dim_from_loop_state_var()
            with tf.control_dependencies(
                [
                    tf.Assert(
                        tf.equal(loop_batch_dim % base_batch_dim, 0),
                        ["loop_batch_dim", loop_batch_dim, "base_batch_dim", base_batch_dim],
                    )
                ]
            ):
                self._parent_tile_multiples = loop_batch_dim // base_batch_dim
        return self._parent_tile_multiples

    def create_state_var(self, name, initial_value=None, data_shape=None):
        """
        A state var is a variable where the initial value is given by the encoder, or a constant,
        and the final value is determined by one step of this rec layer (usually called the decoder).

        :param str name:
        :param tf.Tensor|None initial_value: assumes batch-major, if data_shape is not given
        :param Data|None data_shape:
        :rtype: RecStepByStepLayer.StateVar
        """
        assert name not in self.state_vars
        assert data_shape is not None or initial_value is not None
        if data_shape is not None:
            assert isinstance(data_shape, Data)
            if data_shape.have_batch_axis() and initial_value is not None:
                assert initial_value.shape.dims[data_shape.batch_dim_axis].value is None
        elif initial_value.shape.ndims == 0:
            data_shape = Data(name=name, batch_dim_axis=None, shape=(), dtype=initial_value.dtype.name)
        else:
            assert initial_value.shape.dims[0].value is None  # first is batch dim
            data_shape = Data(
                name=name, batch_dim_axis=0, shape=initial_value.shape.as_list()[1:], dtype=initial_value.dtype.name
            )
        if initial_value is not None:
            # initial_value might have dim 1 in variable dimensions (which are not the batch-dim-axis),
            # see get_rec_initial_output, which should be fine for broadcasting.
            initial_value.set_shape(data_shape.batch_shape)
        var = self.StateVar(parent=self, name=name, initial_value=initial_value, data_shape=data_shape)
        self.state_vars[name] = var
        return var

    def set_state_var_final_value(self, name, final_value):
        """
        :param str name:
        :param tf.Tensor final_value:
        """
        self.state_vars[name].set_final_value(final_value)

    def create_state_vars_recursive(self, name_prefix, initial_values, data_shapes=None):
        """
        :param str|tuple[str] name_prefix: single or same structure as initial_values
        :param T initial_values:
        :param data_shapes: same structure as initial_values or None, but values are of instance :class:`Data`
        :return: same as initial_values, but the state vars
        """

        def _map_state_vars(path, name, initial_value, data_shape):
            assert isinstance(initial_value, tf.Tensor)
            assert data_shape is None or isinstance(data_shape, Data)
            if not name:
                name = "/".join(map(str, (name_prefix,) + path))
            return self.create_state_var(name=name, initial_value=initial_value, data_shape=data_shape)

        if isinstance(name_prefix, str):
            name_per_entry = nest.map_structure(lambda v: None, initial_values)
        else:
            name_per_entry = name_prefix
        if data_shapes is None:
            data_shapes = nest.map_structure(lambda v: None, initial_values)
        nest.assert_same_structure(initial_values, name_per_entry)
        nest.assert_same_structure(initial_values, data_shapes)
        return nest.map_structure_with_tuple_paths(_map_state_vars, name_per_entry, initial_values, data_shapes)

    def assign_state_vars_recursive_flatten(self, name_prefix, values):
        """
        :param str|tuple[str] name_prefix:
        :param T values:
        :rtype: list[tf.Operation]
        """

        def _map_state_vars(path, name, value):
            assert isinstance(value, tf.Tensor)
            if not name:
                name = "/".join(map(str, (name_prefix,) + path))
            return self.state_vars[name].assign(value)

        if isinstance(name_prefix, str):
            name_per_entry = nest.map_structure(lambda v: None, values)
        else:
            name_per_entry = name_prefix
        nest.assert_same_structure(values, name_per_entry)
        return nest.flatten(nest.map_structure_with_tuple_paths(_map_state_vars, name_per_entry, values))

    def have_base_state_vars(self):
        """
        :rtype: bool
        """
        for name, _ in sorted(self.state_vars.items()):
            assert isinstance(name, str)
            if name.startswith("base_"):
                return True
        return False

    def get_batch_dim_from_base_state_var(self):
        """
        :return: batch-dim, from some (any) base state var, scalar, int32
        :rtype: tf.Tensor|int
        """
        for name, v in sorted(self.state_vars.items()):
            assert isinstance(name, str)
            assert isinstance(v, RecStepByStepLayer.StateVar)
            if name.startswith("base_"):
                if v.var_data_shape.have_batch_axis():
                    with tf_util.reuse_name_scope("batch_dim_from_base_state_var_%s" % v.name, absolute=True):
                        return tf.shape(v.var.value())[v.var_data_shape.batch_dim_axis]
        raise Exception("None of the base state vars do have a batch-dim: %s" % self.state_vars)

    def get_batch_dim_from_loop_state_var(self):
        """
        :return: batch-dim, from some (any) loop state var, scalar, int32
        :rtype: tf.Tensor|int
        """
        for name, v in sorted(self.state_vars.items()):
            assert isinstance(v, RecStepByStepLayer.StateVar)
            if not name.startswith("base_") and not name.startswith("stochastic_var_"):
                if v.var_data_shape.batch_dim_axis is not None:
                    with tf_util.reuse_name_scope("batch_dim_from_state_%s" % v.name, absolute=True):
                        return tf.shape(v.var.value())[v.var_data_shape.batch_dim_axis]
        raise Exception("None of the loop state vars do have a batch-dim: %s" % self.state_vars)

    class StochasticVar:
        """
        Manages a stochastic variable, which corresponds to a :class:`ChoiceLayer`.
        """

        def __init__(self, parent_rec_layer, layer_template):
            """
            :param RecStepByStepLayer parent_rec_layer:
            :param _TemplateLayer layer_template:
            """
            self.parent_rec_layer = parent_rec_layer
            self.name = layer_template.name
            self.choice_layer_opts = layer_template.kwargs
            # for compatibility: one way to set the ngram label context via additional choices
            # but older history is not score-dependent stochastic_var anymore
            self.score_dependent = self.choice_layer_opts.get("score_dependent", True)
            self.score_state_var = None
            if self.score_dependent:
                sources = layer_template.kwargs["sources"]
                assert len(sources) == 1
                (source_template,) = sources
                self.score_state_var = self.parent_rec_layer.create_state_var(
                    name="stochastic_var_scores_%s" % self.name, data_shape=source_template.output
                )
            self.choice_state_var = self.parent_rec_layer.create_state_var(
                name="stochastic_var_choice_%s" % self.name, data_shape=layer_template.output
            )

        def assign_score(self, source):
            """
            :param LayerBase source:
            """
            assert self.score_dependent
            assert source.output.is_batch_major and len(source.output.shape) == 1
            scores_in = source.output.placeholder
            input_type = self.choice_layer_opts.get("input_type", "prob")
            if input_type == "prob":
                if source.output_before_activation:
                    scores_in = source.output_before_activation.get_log_output()
                else:
                    from returnn.tf.util.basic import safe_log

                    scores_in = safe_log(scores_in)
            elif input_type == "log_prob":
                pass
            else:
                raise ValueError("Not handled input_type %r" % (input_type,))
            self.score_state_var.set_final_value(final_value=scores_in)

        def get_choice(self):
            """
            :return: the choice value
            :rtype: tf.Tensor
            """
            return self.choice_state_var.read()

    def add_stochastic_var(self, name):
        """
        :param str name:
        :rtype: RecStepByStepLayer.StochasticVar
        """
        assert name not in self.stochastic_vars
        cell = self.cell
        assert isinstance(cell, SubnetworkRecCellSingleStep)
        self.stochastic_vars[name] = self.StochasticVar(self, cell.layer_data_templates[name])
        return self.stochastic_vars[name]

    @classmethod
    def transform_config_dict(cls, d, network, get_layer):
        """
        :param dict[str] d: will modify inplace
        :param returnn.tf.network.TFNetwork network:
        :param ((str) -> LayerBase) get_layer: function to get or construct another layer
        """
        d["_orig_sources"] = d.get("from", "data")  # used for potential custom get_parent_layer of SubnetworkRecCell
        super(RecStepByStepLayer, cls).transform_config_dict(d, network=network, get_layer=get_layer)


class ChoiceStateVarLayer(LayerBase):
    """
    Like :class:`ChoiceLayer`, but we don't do the search/choice ourselves,
    instead we store the scores in a variable, and the final result is another variable,
    which is expected to be set externally.
    This is expected to be used together with :class:`RecStepByStepLayer`.
    """

    layer_class = "choice_state_var"

    # noinspection PyUnusedLocal
    def __init__(
        self,
        beam_size,
        search=NotSpecified,
        input_type="prob",
        prob_scale=1.0,
        base_beam_score_scale=1.0,
        random_sample_scale=0.0,
        length_normalization=True,
        custom_score_combine=None,
        source_beam_sizes=None,
        scheduled_sampling=False,
        cheating=False,
        explicit_search_sources=None,
        score_dependent=True,
        **kwargs,
    ):
        super(ChoiceStateVarLayer, self).__init__(**kwargs)
        rec_layer = self.network.parent_layer
        assert isinstance(rec_layer, RecStepByStepLayer)
        self.stochastic_var = rec_layer.stochastic_vars[self.name]
        cell = rec_layer.cell
        assert isinstance(cell, SubnetworkRecCellSingleStep)
        assert self.network in {cell.net, cell.net_delayed_update}
        if self.network == cell.net:
            rec_layer.stochastic_var_order.append(self.name)
        # for compatibility: one way to set the ngram label context via additional choices
        # but older history is not score-dependent stochastic_var anymore
        self.score_dependent = score_dependent
        if self.score_dependent:
            assert len(self.sources) == 1
            (source,) = self.sources
            self.stochastic_var.assign_score(source)
        self.output.placeholder = self.stochastic_var.get_choice()

    @classmethod
    def transform_config_dict(cls, d, network, get_layer):
        """
        :param dict[str] d: will modify inplace
        :param returnn.tf.network.TFNetwork network:
        :param ((str) -> LayerBase) get_layer: function to get or construct another layer
        """
        assert d.get("from", NotSpecified) is not NotSpecified, "specify 'from' explicitly for choice layer"
        if not isinstance(d["from"], (tuple, list)):
            d["from"] = [d["from"]]
        if d.get("target", NotSpecified) is not None:
            assert "target" in d, "%s: specify 'target' explicitly" % (cls.__name__,)
            if isinstance(d["target"], str):
                d["target"] = [d["target"]]
            assert isinstance(d["target"], list)
            assert len(d["target"]) == len(d["from"])
        if d.get("explicit_search_source"):
            assert "explicit_search_sources" not in d
            d["explicit_search_sources"] = [get_layer(d.pop("explicit_search_source"))]
        elif d.get("explicit_search_sources"):
            assert isinstance(d["explicit_search_sources"], (list, tuple))
            d["explicit_search_sources"] = [get_layer(name) for name in d["explicit_search_sources"]]
        parent_rec_layer = network.parent_layer
        if parent_rec_layer:  # might be None at first template construction
            assert isinstance(parent_rec_layer, RecStepByStepLayer)
            cell = parent_rec_layer.cell
            assert isinstance(cell, SubnetworkRecCellSingleStep)
            assert network in {cell.net, cell.net_delayed_update}
            if network == cell.net_delayed_update:
                # Remove the dependency to the scores.
                d["from"] = ()
                d["score_dependent"] = False
        super(ChoiceStateVarLayer, cls).transform_config_dict(d, network=network, get_layer=get_layer)

    @classmethod
    def get_out_data_from_opts(cls, name, sources, target, network, **kwargs):
        """
        :param str name:
        :param list[LayerBase] sources:
        :param str target:
        :param returnn.tf.network.TFNetwork network:
        :rtype: Data
        """
        # This is simplified from ChoiceLayer.get_out_data_from_opts.
        target = target[0] if isinstance(target, list) else target  # only the first matters here
        if target:
            out_data = cls._static_get_target_value(
                target=target, network=network, mark_data_key_as_used=False
            ).copy_template(name="%s_output" % name)
            out_data.available_for_inference = True  # in inference, we would do search
        else:  # no target. i.e. we must do search
            # Output will be the sparse version of the input.
            out_data = sources[0].output.copy_template().copy_as_batch_major()
            dim_tags = list(out_data.dim_tags)
            del dim_tags[out_data.feature_dim_axis]
            out_data = Data(
                name="%s_output" % name,
                dim_tags=dim_tags,
                sparse=True,
                dim=out_data.dim,
                batch=out_data.batch.copy_set_beam(None) if out_data.batch else network.get_global_batch_info(),
            )
        return out_data


def main(argv):
    """
    Main entry.
    """
    argparser = argparse.ArgumentParser(description="Compile some op")
    argparser.add_argument("config", help="filename to config-file")
    argparser.add_argument(
        "--epoch", type=int, default=None, help="specific epoch to construct, use for dynamic network definitions"
    )
    argparser.add_argument("--train", type=int, default=0, help="0 disable (default), 1 enable, -1 dynamic")
    argparser.add_argument("--eval", type=int, default=0, help="calculate losses. 0 disable (default), 1 enable")
    argparser.add_argument("--search", type=int, default=0, help="beam search. 0 disable (default), 1 enable")
    argparser.add_argument(
        "--device", default="cpu", help="'cpu' (default) or 'gpu'. optimizes the graph for this device"
    )
    argparser.add_argument("--verbosity", default=4, type=int, help="5 for all seqs (default: 4)")
    argparser.add_argument("--summaries_tensor_name", help="create Tensor for tf.compat.v1.summary.merge_all()")
    argparser.add_argument("--rec_step_by_step", help="make step-by-step graph for this rec layer (eg. 'output')")
    argparser.add_argument("--rec_step_by_step_output_file", help="store meta info for rec_step_by_step (JSON)")
    argparser.add_argument("--output_file", help="allowed extensions: pb, pbtxt, meta, metatxt, logdir")
    argparser.add_argument("--output_file_model_params_list", help="line-based, names of model params")
    argparser.add_argument("--output_file_state_vars_list", help="line-based, name of state vars")
    argparser.add_argument("--update_i_in_graph", action="store_true", help="whether to update i in the graph")
    args = argparser.parse_args(argv[1:])
    assert args.train in [0, 1, -1] and args.eval in [0, 1] and args.search in [0, 1]
    init(config_filename=args.config, log_verbosity=args.verbosity, device=args.device)

    if args.epoch and "get_network" in config.typed_dict:
        net_dict = config.typed_dict["get_network"](epoch=args.epoch)
    elif args.epoch and "pretrain" in config.typed_dict:
        raise NotImplementedError(
            "Compiling a network at a specific epoch using pre-train logic is not implemented yet"
        )
    else:
        assert "network" in config.typed_dict
        net_dict = config.typed_dict["network"]
    assert isinstance(net_dict, dict), "network should return dict but is %s" % type(net_dict)

    if args.rec_step_by_step:
        RecStepByStepLayer.prepare_compile(
            rec_layer_name=args.rec_step_by_step, net_dict=net_dict, opts={"update_i_in_graph": args.update_i_in_graph}
        )
    with tf.Graph().as_default() as graph:
        assert isinstance(graph, tf.Graph)
        print("Create graph...")
        # See :func:`Engine._init_network`.
        tf_compat.v1.set_random_seed(42)
        if args.train < 0:
            from returnn.tf.util.basic import get_global_train_flag_placeholder

            train_flag = get_global_train_flag_placeholder()
        else:
            train_flag = bool(args.train)
        eval_flag = bool(args.eval)
        search_flag = bool(args.search)
        network = create_graph(train_flag=train_flag, eval_flag=eval_flag, search_flag=search_flag, net_dict=net_dict)

        if args.rec_step_by_step:
            RecStepByStepLayer.post_compile(
                rec_layer_name=args.rec_step_by_step,
                network=network,
                output_file_name=args.rec_step_by_step_output_file,
            )

        from returnn.tf.layers.base import LayerBase

        for layer in network.layers.values():
            assert isinstance(layer, LayerBase)
            if layer.output.time_dim_axis is None:
                continue
            if layer.output.batch_dim_axis is None:
                continue
            with tf_util.reuse_name_scope(layer.get_absolute_name_scope_prefix()[:-1], absolute=True):
                out = layer.output.copy_as_batch_major()
                if out.have_feature_axis():
                    out = out.copy_with_feature_last()
                tf.identity(out.placeholder, name="output_batch_major")

        tf.group(*network.get_post_control_dependencies(), name="post_control_dependencies")

        # Do some cleanup of collections which do not contain tensors or operations,
        # because the tf.train.import_meta_graph code might fail otherwise.
        tf_compat.v1.get_collection_ref(CollectionKeys.RETURNN_LAYERS).clear()

        if args.summaries_tensor_name:
            summaries_tensor = tf_compat.v1.summary.merge_all()
            assert isinstance(summaries_tensor, tf.Tensor), "no summaries in the graph?"
            tf.identity(summaries_tensor, name=args.summaries_tensor_name)

        if args.output_file and os.path.splitext(args.output_file)[1] in [".meta", ".metatxt"]:
            # https://www.tensorflow.org/api_guides/python/meta_graph
            saver = tf_compat.v1.train.Saver(var_list=network.get_saveable_params_list(), max_to_keep=2**31 - 1)
            graph_def = saver.export_meta_graph()
        else:
            graph_def = graph.as_graph_def(add_shapes=True)

        print("Graph collection keys:", graph.get_all_collection_keys())
        print("Graph num operations:", len(graph.get_operations()))
        print("Graph def size:", util.human_bytes_size(graph_def.ByteSize()))

        if args.output_file:
            filename = args.output_file
            _, ext = os.path.splitext(filename)
            if ext == ".logdir":
                print("Write TF events to logdir:", filename)
                writer = tf_compat.v1.summary.FileWriter(logdir=filename)
                writer.add_graph(graph)
                writer.flush()
            else:
                assert ext in [".pb", ".pbtxt", ".meta", ".metatxt"], "filename %r extension invalid" % filename
                print("Write graph to file:", filename)
                graph_io.write_graph(
                    graph_def,
                    logdir=os.path.dirname(filename),
                    name=os.path.basename(filename),
                    as_text=ext.endswith("txt"),
                )
        else:
            print("Use --output_file if you want to store the graph.")

        if args.output_file_model_params_list:
            print("Write model param list to:", args.output_file_model_params_list)
            with open(args.output_file_model_params_list, "w") as f:
                for param in network.get_params_list():
                    assert param.name[-2:] == ":0"
                    f.write("%s\n" % param.name[:-2])

        if args.output_file_state_vars_list:
            print("Write state var list to:", args.output_file_state_vars_list)
            with open(args.output_file_state_vars_list, "w") as f:
                for param in tf_compat.v1.get_collection(CollectionKeys.STATE_VARS):
                    assert param.name[-2:] == ":0"
                    f.write("%s\n" % param.name[:-2])


if __name__ == "__main__":
    main(sys.argv)



# -*- coding: utf-8 -*-
# Copyright (c) 2011-2021, Albert Zeyer, www.az2000.de
# All rights reserved.
# file created 2011-04-15


# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
https://github.com/albertz/py_better_exchook

This is a simple replacement for the standard Python exception handler (sys.excepthook).
In addition to what the standard handler does, it also prints all referenced variables
(no matter if local, global or builtin) of the code line of each stack frame.
See below for some examples and some example output.

See these functions:

- better_exchook
- format_tb / print_tb
- iter_traceback
- get_current_frame
- dump_all_thread_tracebacks
- install
- setup_all
- replace_traceback_format_tb
- replace_traceback_print_tb

Although there might be a few more useful functions, thus we export all of them.

Also see the demo/tests at the end.
"""

from __future__ import annotations

import sys
import os
import os.path
import threading
import keyword
import inspect
import contextlib

try:
    import typing
except ImportError:
    typing = None

try:
    from traceback import StackSummary, FrameSummary
except ImportError:

    class _Dummy:
        pass

    StackSummary = FrameSummary = _Dummy

# noinspection PySetFunctionToLiteral,SpellCheckingInspection
py_keywords = set(keyword.kwlist) | set(["None", "True", "False"])

_cur_pwd = os.getcwd()
_threading_main_thread = threading.main_thread() if hasattr(threading, "main_thread") else None

try:
    # noinspection PyUnresolvedReferences,PyUnboundLocalVariable
    unicode
except NameError:  # Python3
    unicode = str  # Python 3 compatibility

try:
    # noinspection PyUnresolvedReferences,PyUnboundLocalVariable
    raw_input
except NameError:  # Python3
    raw_input = input


PY3 = sys.version_info[0] >= 3


def parse_py_statement(line):
    """
    Parse Python statement into tokens.
    Note that this is incomplete.
    It should be simple and fast and just barely enough for what we need here.

    Reference:
    https://docs.python.org/3/reference/lexical_analysis.html

    :param str line:
    :return: yields (type, value)
    :rtype: typing.Iterator[typing.Tuple[str,str]]
    """
    state = 0
    cur_token = ""
    spaces = " \t\n"
    ops = ".,;:+-*/%&!=|(){}[]^<>"
    i = 0

    def _escape_char(_c):
        if _c == "n":
            return "\n"
        elif _c == "t":
            return "\t"
        else:
            return _c

    while i < len(line):
        c = line[i]
        i += 1
        if state == 0:
            if c in spaces:
                pass
            elif c in ops:
                yield "op", c
            elif c == "#":
                state = 6
            elif c == '"':
                state = 1
            elif c == "'":
                state = 2
            else:
                cur_token = c
                state = 3
        elif state == 1:  # string via "
            if c == "\\":
                state = 4
            elif c == '"':
                yield "str", cur_token
                cur_token = ""
                state = 0
            else:
                cur_token += c
        elif state == 2:  # string via '
            if c == "\\":
                state = 5
            elif c == "'":
                yield "str", cur_token
                cur_token = ""
                state = 0
            else:
                cur_token += c
        elif state == 3:  # identifier
            if c in spaces + ops + "#":
                yield "id", cur_token
                cur_token = ""
                state = 0
                i -= 1
            elif c == '"':  # identifier is string prefix
                cur_token = ""
                state = 1
            elif c == "'":  # identifier is string prefix
                cur_token = ""
                state = 2
            else:
                cur_token += c
        elif state == 4:  # escape in "
            cur_token += _escape_char(c)
            state = 1
        elif state == 5:  # escape in '
            cur_token += _escape_char(c)
            state = 2
        elif state == 6:  # comment
            cur_token += c
    if state == 3:
        yield "id", cur_token
    elif state == 6:
        yield "comment", cur_token


def parse_py_statements(source_code):
    """
    :param str source_code:
    :return: via :func:`parse_py_statement`
    :rtype: typing.Iterator[typing.Tuple[str,str]]
    """
    for line in source_code.splitlines():
        for t in parse_py_statement(line):
            yield t


def grep_full_py_identifiers(tokens):
    """
    :param typing.Iterable[(str,str)] tokens:
    :rtype: typing.Iterator[str]
    """
    global py_keywords
    tokens = list(tokens)
    i = 0
    while i < len(tokens):
        token_type, token = tokens[i]
        i += 1
        if token_type != "id":
            continue
        while i + 1 < len(tokens) and tokens[i] == ("op", ".") and tokens[i + 1][0] == "id":
            token += "." + tokens[i + 1][1]
            i += 2
        if token == "":
            continue
        if token in py_keywords:
            continue
        if token[0] in ".0123456789":
            continue
        yield token


def set_linecache(filename, source):
    """
    The :mod:`linecache` module has some cache of the source code for the current source.
    Sometimes it fails to find the source of some files.
    We can explicitly set the source for some filename.

    :param str filename:
    :param str source:
    :return: nothing
    """
    import linecache

    linecache.cache[filename] = None, None, [line + "\n" for line in source.splitlines()], filename


# noinspection PyShadowingBuiltins
def simple_debug_shell(globals, locals):
    """
    :param dict[str,typing.Any] globals:
    :param dict[str,typing.Any] locals:
    :return: nothing
    """
    try:
        import readline
    except ImportError:
        pass  # ignore
    compile_string_fn = "<simple_debug_shell input>"
    while True:
        try:
            s = raw_input("> ")
        except (KeyboardInterrupt, EOFError):
            print("breaked debug shell: " + sys.exc_info()[0].__name__)
            break
        if s.strip() == "":
            continue
        try:
            c = compile(s, compile_string_fn, "single")
        except Exception as e:
            print("%s : %s in %r" % (e.__class__.__name__, str(e), s))
        else:
            set_linecache(compile_string_fn, s)
            # noinspection PyBroadException
            try:
                ret = eval(c, globals, locals)
            except (KeyboardInterrupt, SystemExit):
                print("debug shell exit: " + sys.exc_info()[0].__name__)
                break
            except Exception:
                print("Error executing %r" % s)
                better_exchook(*sys.exc_info(), autodebugshell=False)
            else:
                # noinspection PyBroadException
                try:
                    if ret is not None:
                        print(ret)
                except Exception:
                    print("Error printing return value of %r" % s)
                    better_exchook(*sys.exc_info(), autodebugshell=False)


# keep non-PEP8 argument name for compatibility
# noinspection PyPep8Naming
def debug_shell(user_ns, user_global_ns, traceback=None, execWrapper=None):
    """
    Spawns some interactive shell. Tries to use IPython if available.
    Falls back to :func:`pdb.post_mortem` or :func:`simple_debug_shell`.

    :param dict[str,typing.Any] user_ns:
    :param dict[str,typing.Any] user_global_ns:
    :param traceback:
    :param execWrapper:
    :return: nothing
    """
    ipshell = None
    try:
        # noinspection PyPackageRequirements
        import IPython

        have_ipython = True
    except ImportError:
        have_ipython = False

    if not ipshell and traceback and have_ipython:
        # noinspection PyBroadException
        try:
            # noinspection PyPackageRequirements,PyUnresolvedReferences
            from IPython.core.debugger import Pdb

            # noinspection PyPackageRequirements,PyUnresolvedReferences
            from IPython.terminal.debugger import TerminalPdb

            # noinspection PyPackageRequirements,PyUnresolvedReferences
            from IPython.terminal.ipapp import TerminalIPythonApp

            ipapp = TerminalIPythonApp.instance()
            ipapp.interact = False  # Avoid output (banner, prints)
            ipapp.initialize(argv=[])
            def_colors = ipapp.shell.colors
            pdb_obj = TerminalPdb(def_colors)
            pdb_obj.botframe = None  # not sure. exception otherwise at quit

            def ipshell():
                """
                Run the IPython shell.
                """
                pdb_obj.interaction(None, traceback)

        except Exception:
            print("IPython Pdb exception:")
            better_exchook(*sys.exc_info(), autodebugshell=False, file=sys.stdout)

    if not ipshell and have_ipython:
        # noinspection PyBroadException
        try:
            # noinspection PyPackageRequirements,PyUnresolvedReferences
            import IPython

            # noinspection PyPackageRequirements,PyUnresolvedReferences
            import IPython.terminal.embed

            class DummyMod:
                """Dummy module"""

            module = DummyMod()
            module.__dict__ = user_global_ns
            module.__name__ = "_DummyMod"
            if "__name__" not in user_ns:
                user_ns = user_ns.copy()
                user_ns["__name__"] = "_DummyUserNsMod"
            ipshell = IPython.terminal.embed.InteractiveShellEmbed.instance(user_ns=user_ns, user_module=module)
        except Exception:
            print("IPython not available:")
            better_exchook(*sys.exc_info(), autodebugshell=False, file=sys.stdout)
        else:
            if execWrapper:
                old = ipshell.run_code
                ipshell.run_code = lambda code: execWrapper(lambda: old(code))
    if ipshell:
        ipshell()
    else:
        print("Use simple pdb debug shell:")
        if traceback:
            import pdb

            pdb.post_mortem(traceback)
        else:
            simple_debug_shell(user_global_ns, user_ns)


def output_limit():
    """
    :return: num chars
    :rtype: int
    """
    return 300


def fallback_findfile(filename):
    """
    :param str filename:
    :return: try to find the full filename, e.g. in modules, etc
    :rtype: str|None
    """
    mods = [m for m in list(sys.modules.values()) if m and getattr(m, "__file__", None) and filename in m.__file__]
    if len(mods) == 0:
        return None
    alt_fn = mods[0].__file__
    if alt_fn[-4:-1] == ".py":
        alt_fn = alt_fn[:-1]  # *.pyc or whatever
    if not os.path.exists(alt_fn) and alt_fn.startswith("./"):
        # Maybe current dir changed.
        alt_fn2 = _cur_pwd + alt_fn[1:]
        if os.path.exists(alt_fn2):
            return alt_fn2
        # Try dirs of some other mods.
        for m in ["__main__", "better_exchook"]:
            if hasattr(sys.modules.get(m), "__file__"):
                alt_fn2 = os.path.dirname(sys.modules[m].__file__) + alt_fn[1:]
                if os.path.exists(alt_fn2):
                    return alt_fn2
    return alt_fn


def is_source_code_missing_brackets(source_code, prioritize_missing_open=False):
    """
    We check whether this source code snippet (e.g. one line) is complete/even w.r.t. opening/closing brackets.

    :param str source_code:
    :param bool prioritize_missing_open: once we found any missing open bracket, directly return -1
    :return: 1 if missing_close, -1 if missing_open, 0 otherwise.
        I.e. whether there are missing open/close brackets.
        E.g. this would mean that you might want to include the prev/next source code line as well in the stack trace.
    :rtype: int
    """
    open_brackets = "[{("
    close_brackets = "]})"
    last_bracket = [-1]  # stack
    counters = [0] * len(open_brackets)
    missing_open = False
    for t_type, t_content in list(parse_py_statements(source_code)):
        if t_type != "op":
            continue  # we are from now on only interested in ops (including brackets)
        if t_content in open_brackets:
            idx = open_brackets.index(t_content)
            counters[idx] += 1
            last_bracket.append(idx)
        elif t_content in close_brackets:
            idx = close_brackets.index(t_content)
            if last_bracket[-1] == idx:
                counters[idx] -= 1
                del last_bracket[-1]
            else:
                if prioritize_missing_open:
                    return -1
                missing_open = True
    missing_close = not all([c == 0 for c in counters])
    if missing_close:
        return 1
    if missing_open:
        return -1
    return 0


def is_source_code_missing_open_brackets(source_code):
    """
    We check whether this source code snippet (e.g. one line) is complete/even w.r.t. opening/closing brackets.

    :param str source_code:
    :return: whether there are missing open brackets.
        E.g. this would mean that you might want to include the previous source code line as well in the stack trace.
    :rtype: bool
    """
    return is_source_code_missing_brackets(source_code, prioritize_missing_open=True) < 0


def get_source_code(filename, lineno, module_globals=None):
    """
    :param str filename:
    :param int lineno:
    :param dict[str,typing.Any]|None module_globals:
    :return: source code of that line (including newline)
    :rtype: str
    """
    import linecache

    linecache.checkcache(filename)
    source_code = linecache.getline(filename, lineno, module_globals)
    # In case of a multi-line statement, lineno is usually the last line.
    # We are checking for missing open brackets and add earlier code lines.
    start_line = end_line = lineno
    lines = None
    while True:
        missing_bracket_level = is_source_code_missing_brackets(source_code)
        if missing_bracket_level == 0:
            break
        if not lines:
            lines = linecache.getlines(filename, module_globals)
        if missing_bracket_level < 0:  # missing open bracket, add prev line
            start_line -= 1
            if start_line < 1:  # 1-indexed
                break
        else:
            end_line += 1
            if end_line > len(lines):  # 1-indexed
                break
        source_code = "".join(lines[start_line - 1 : end_line])  # 1-indexed
    return source_code


def str_visible_len(s):
    """
    :param str s:
    :return: len without escape chars
    :rtype: int
    """
    import re

    # via: https://github.com/chalk/ansi-regex/blob/master/index.js
    s = re.sub("[\x1b\x9b][\\[()#;?]*(?:[0-9]{1,4}(?:;[0-9]{0,4})*)?[0-9A-PRZcf-nqry=><]", "", s)
    return len(s)


def add_indent_lines(prefix, s):
    """
    :param str prefix:
    :param str s:
    :return: s with prefix indent added to all lines
    :rtype: str
    """
    if not s:
        return prefix
    prefix_len = str_visible_len(prefix)
    lines = s.splitlines(True)
    return "".join([prefix + lines[0]] + [" " * prefix_len + line for line in lines[1:]])


def get_indent_prefix(s):
    """
    :param str s:
    :return: the indent spaces of s
    :rtype: str
    """
    return s[: len(s) - len(s.lstrip())]


def get_same_indent_prefix(lines):
    """
    :param list[] lines:
    :rtype: str|None
    """
    if not lines:
        return ""
    prefix = get_indent_prefix(lines[0])
    if not prefix:
        return ""
    if all([line.startswith(prefix) for line in lines]):
        return prefix
    return None


def remove_indent_lines(s):
    """
    :param str s:
    :return: remove as much indentation as possible
    :rtype: str
    """
    if not s:
        return ""
    lines = s.splitlines(True)
    prefix = get_same_indent_prefix(lines)
    if prefix is None:  # not in expected format. just lstrip all lines
        return "".join([line.lstrip() for line in lines])
    return "".join([line[len(prefix) :] for line in lines])


def replace_tab_indent(s, replace="    "):
    """
    :param str s: string with tabs
    :param str replace: e.g. 4 spaces
    :rtype: str
    """
    prefix = get_indent_prefix(s)
    return prefix.replace("\t", replace) + s[len(prefix) :]


def replace_tab_indents(s, replace="    "):
    """
    :param str s: multi-line string with tabs
    :param str replace: e.g. 4 spaces
    :rtype: str
    """
    lines = s.splitlines(True)
    return "".join([replace_tab_indent(line, replace) for line in lines])


def to_bool(s, fallback=None):
    """
    :param str s: str to be converted to bool, e.g. "1", "0", "true", "false"
    :param T fallback: if s is not recognized as a bool
    :return: boolean value, or fallback
    :rtype: bool|T
    """
    if not s:
        return fallback
    s = s.lower()
    if s in ["1", "true", "yes", "y"]:
        return True
    if s in ["0", "false", "no", "n"]:
        return False
    return fallback


class Color:
    """
    Helper functions provided to perform terminal coloring.
    """

    ColorIdxTable = {
        k: i for (i, k) in enumerate(["black", "red", "green", "yellow", "blue", "magenta", "cyan", "white"])
    }

    @classmethod
    def get_global_color_enabled(cls):
        """
        :rtype: bool
        """
        return to_bool(os.environ.get("CLICOLOR", ""), fallback=True)

    @classmethod
    def is_dark_terminal_background(cls):
        """
        :return: Whether we have a dark Terminal background color, or None if unknown.
            We currently just check the env var COLORFGBG,
            which some terminals define like "<foreground-color>:<background-color>",
            and if <background-color> in {0,1,2,3,4,5,6,8}, then we have some dark background.
            There are many other complex heuristics we could do here, which work in some cases but not in others.
            See e.g. `here <https://stackoverflow.com/questions/2507337/terminals-background-color>`__.
            But instead of adding more heuristics, we think that explicitly setting COLORFGBG would be the best thing,
            in case it's not like you want it.
        :rtype: bool|None
        """
        if os.environ.get("COLORFGBG", None):
            parts = os.environ["COLORFGBG"].split(";")
            try:
                last_number = int(parts[-1])
                if 0 <= last_number <= 6 or last_number == 8:
                    return True
                else:
                    return False
            except ValueError:  # not an integer?
                pass
        return None  # unknown (and bool(None) == False, i.e. expect light by default)

    def __init__(self, enable=None):
        """
        :param bool|None enable:
        """
        if enable is None:
            enable = self.get_global_color_enabled()
        self.enable = enable
        self._dark_terminal_background = self.is_dark_terminal_background()
        # Set color palettes (will be used sometimes as bold, sometimes as normal).
        # 5 colors, for: code/general, error-msg, string, comment, line-nr.
        # Try to set them in a way such that if we guessed the terminal background color wrongly,
        # it is still not too bad (although people might disagree here...).
        if self._dark_terminal_background:
            self.fg_colors = ["yellow", "red", "cyan", "white", "magenta"]
        else:
            self.fg_colors = ["blue", "red", "cyan", "white", "magenta"]

    def color(self, s, color=None, bold=False):
        """
        :param str s:
        :param str|None color: sth in self.ColorIdxTable
        :param bool bold:
        :return: s optionally wrapped with ansi escape codes
        :rtype: str
        """
        if not self.enable:
            return s
        code_seq = []
        if color:
            code_seq += [30 + self.ColorIdxTable[color]]  # foreground color
        if bold:
            code_seq += [1]
        if not code_seq:
            return s
        start = "\x1b[%sm" % ";".join(map(str, code_seq))
        end = "\x1b[0m"
        while s[:1] == " ":  # move prefix spaces outside
            start = " " + start
            s = s[1:]
        while s[-1:] == " ":  # move postfix spaces outside
            end += " "
            s = s[:-1]
        return start + s + end

    def __call__(self, *args, **kwargs):
        return self.color(*args, **kwargs)

    def py_syntax_highlight(self, s):
        """
        :param str s:
        :rtype: str
        """
        if not self.enable:
            return s
        state = 0
        spaces = " \t\n"
        ops = ".,;:+-*/%&!=|(){}[]^<>"
        i = 0
        cur_token = ""
        # pos in s -> color kwargs
        color_args = {0: {}, len(s): {}}  # type: typing.Dict[int,typing.Dict[str,typing.Any]]

        def finish_identifier():
            """
            Reset color to standard for current identifier.
            """
            if cur_token in py_keywords:
                color_args[max([k for k in color_args.keys() if k < i])] = {"color": self.fg_colors[0]}

        while i < len(s):
            c = s[i]
            i += 1
            if c == "\n":
                if state == 3:
                    finish_identifier()
                color_args[i] = {}
                state = 0
            elif state == 0:
                if c in spaces:
                    pass
                elif c in ops:
                    color_args[i - 1] = {"color": self.fg_colors[0]}
                    color_args[i] = {}
                elif c == "#":
                    color_args[i - 1] = {"color": self.fg_colors[3]}
                    state = 6
                elif c == '"':
                    color_args[i - 1] = {"color": self.fg_colors[2]}
                    state = 1
                elif c == "'":
                    color_args[i - 1] = {"color": self.fg_colors[2]}
                    state = 2
                else:
                    cur_token = c
                    color_args[i - 1] = {}
                    state = 3
            elif state == 1:  # string via "
                if c == "\\":
                    state = 4
                elif c == '"':
                    color_args[i] = {}
                    state = 0
            elif state == 2:  # string via '
                if c == "\\":
                    state = 5
                elif c == "'":
                    color_args[i] = {}
                    state = 0
            elif state == 3:  # identifier
                if c in spaces + ops + "#\"'":
                    finish_identifier()
                    color_args[i] = {}
                    state = 0
                    i -= 1
                else:
                    cur_token += c
            elif state == 4:  # escape in "
                state = 1
            elif state == 5:  # escape in '
                state = 2
            elif state == 6:  # comment
                pass
        if state == 3:
            finish_identifier()
        out = ""
        i = 0
        while i < len(s):
            j = min([k for k in color_args.keys() if k > i])
            out += self.color(s[i:j], **color_args[i])
            i = j
        return out


class DomTerm:
    """
    DomTerm (https://github.com/PerBothner/DomTerm/) is a terminal emulator
    with many extended escape codes, such as folding text away, or even generic HTML.
    We can make use of some of these features (currently just folding text).
    """

    _is_domterm = None

    @classmethod
    def is_domterm(cls):
        """
        :return: whether we are inside DomTerm
        :rtype: bool
        """
        import os

        if cls._is_domterm is not None:
            return cls._is_domterm
        if not os.environ.get("DOMTERM"):
            cls._is_domterm = False
            return False
        cls._is_domterm = True
        return True

    @contextlib.contextmanager
    def logical_block(self, file=sys.stdout):
        """
        :param io.TextIOBase|io.StringIO file:
        """
        file.write("\033]110\007")
        yield
        file.write("\033]111\007")

    @contextlib.contextmanager
    def hide_button_span(self, mode, file=sys.stdout):
        """
        :param int mode: 1 or 2
        :param io.TextIOBase|io.StringIO file:
        """
        file.write("\033[83;%iu" % mode)
        yield
        file.write("\033[83;0u")

    # noinspection PyMethodMayBeStatic
    def indentation(self, file=sys.stdout):
        """
        :param io.TextIOBase|io.StringIO file:
        """
        file.write('\033]114;"│"\007')

    # noinspection PyMethodMayBeStatic
    def hide_button(self, file=sys.stdout):
        """
        :param io.TextIOBase|io.StringIO file:
        """
        file.write("\033[16u▶▼\033[17u")

    @contextlib.contextmanager
    def _temp_replace_attrib(self, obj, attr, new_value):
        old_value = getattr(obj, attr)
        setattr(obj, attr, new_value)
        yield old_value
        setattr(obj, attr, old_value)

    @contextlib.contextmanager
    def fold_text_stream(self, prefix, postfix="", hidden_stream=None, **kwargs):
        """
        :param str prefix: always visible
        :param str postfix: always visible, right after.
        :param io.TextIOBase|io.StringIO hidden_stream: sys.stdout by default.
            If this is sys.stdout, it will replace that stream,
            and collect the data during the context (in the `with` block).
        """
        import io

        if hidden_stream is None:
            hidden_stream = sys.stdout
        assert isinstance(hidden_stream, io.IOBase)
        assert hidden_stream is sys.stdout, "currently not supported otherwise"
        hidden_buf = io.StringIO()
        with self._temp_replace_attrib(sys, "stdout", hidden_buf):
            yield
        self.fold_text(prefix=prefix, postfix=postfix, hidden=hidden_buf.getvalue(), **kwargs)

    def fold_text(self, prefix, hidden, postfix="", file=None, align=0):
        """
        :param str prefix: always visible
        :param str hidden: hidden
            If this is sys.stdout, it will replace that stream,
            and collect the data during the context (in the `with` block).
        :param str postfix: always visible, right after. "" by default.
        :param io.TextIOBase|io.StringIO file: sys.stdout by default.
        :param int align: remove this number of initial chars from hidden
        """
        if file is None:
            file = sys.stdout
        # Extra logic: Multi-line hidden. Add initial "\n" if not there.
        if "\n" in hidden:
            if hidden[:1] != "\n":
                hidden = "\n" + hidden
        # Extra logic: A final "\n" of hidden, make it always visible such that it looks nicer.
        if hidden[-1:] == "\n":
            hidden = hidden[:-1]
            postfix += "\n"
        if self.is_domterm():
            with self.logical_block(file=file):
                self.indentation(file=file)
                self.hide_button(file=file)
                file.write(prefix)
                if prefix.endswith("\x1b[0m"):
                    file.write(" ")  # bug in DomTerm?
                with self.hide_button_span(2, file=file):
                    hidden_ls = hidden.split("\n")
                    hidden_ls = [s[align:] for s in hidden_ls]
                    hidden = "\033]118\007".join(hidden_ls)
                    file.write(hidden)
        else:
            file.write(prefix)
            file.write(hidden.replace("\n", "\n "))
        file.write(postfix)
        file.flush()

    def fold_text_string(self, prefix, hidden, **kwargs):
        """
        :param str prefix:
        :param str hidden:
        :param kwargs: passed to :func:`fold_text`
        :rtype: str
        """
        import io

        output_buf = io.StringIO()
        self.fold_text(prefix=prefix, hidden=hidden, file=output_buf, **kwargs)
        return output_buf.getvalue()


def is_at_exit():
    """
    Some heuristics to figure out whether this is called at a stage where the Python interpreter is shutting down.

    :return: whether the Python interpreter is currently in the process of shutting down
    :rtype: bool
    """
    if _threading_main_thread is not None:
        if not hasattr(threading, "main_thread"):
            return True
        if threading.main_thread() != _threading_main_thread:
            return True
        if not _threading_main_thread.is_alive():
            return True
    return False


class _OutputLinesCollector:
    def __init__(self, color):
        """
        :param Color color:
        """
        self.color = color
        self.lines = []
        self.dom_term = DomTerm() if DomTerm.is_domterm() else None

    def __call__(self, s1, s2=None, **kwargs):
        """
        Adds to self.lines.
        This strange function signature is for historical reasons.

        :param str s1:
        :param str|None s2:
        :param kwargs: passed to self.color
        """
        if kwargs:
            s1 = self.color(s1, **kwargs)
        if s2 is not None:
            s1 = add_indent_lines(s1, s2)
        self.lines.append(s1 + "\n")

    @contextlib.contextmanager
    def fold_text_ctx(self, line):
        """
        Folds text, via :class:`DomTerm`, if available.
        Notes that this temporarily overwrites self.lines.

        :param str line: always visible
        """
        if not self.dom_term:
            self.__call__(line)
            yield
            return
        self.lines, old_lines = [], self.lines  # overwrite self.lines
        yield  # collect output (in new self.lines)
        self.lines, new_lines = old_lines, self.lines  # recover self.lines
        hidden_text = "".join(new_lines)
        import io

        output_buf = io.StringIO()
        prefix = ""
        while line[:1] == " ":
            prefix += " "
            line = line[1:]
        self.dom_term.fold_text(line, hidden=hidden_text, file=output_buf, align=len(prefix))
        output_text = prefix[1:] + output_buf.getvalue()
        self.lines.append(output_text)

    def _pp_extra_info(self, obj, depth_limit=3):
        """
        :param typing.Any|typing.Sized obj:
        :param int depth_limit:
        :rtype: str
        """
        # We want to exclude array types from Numpy, TensorFlow, PyTorch, etc.
        # Getting __getitem__ or __len__ on them even could lead to unexpected results or deadlocks,
        # depending on the context (e.g. inside a TF session run, extending the graph is unexpected).
        if hasattr(obj, "shape"):
            return ""
        s = []
        if hasattr(obj, "__len__"):
            # noinspection PyBroadException
            try:
                if type(obj) in (str, unicode, list, tuple, dict) and len(obj) <= 5:
                    pass  # don't print len in this case
                else:
                    s += ["len = " + str(obj.__len__())]
            except Exception:
                pass
        if depth_limit > 0 and hasattr(obj, "__getitem__"):
            # noinspection PyBroadException
            try:
                if type(obj) in (str, unicode):
                    pass  # doesn't make sense to get subitems here
                else:
                    subobj = obj.__getitem__(0)  # noqa
                    extra_info = self._pp_extra_info(subobj, depth_limit - 1)
                    if extra_info != "":
                        s += ["_[0]: {" + extra_info + "}"]
            except Exception:
                pass
        return ", ".join(s)

    def pretty_print(self, obj):
        """
        :param typing.Any obj:
        :rtype: str
        """
        s = repr(obj)
        limit = output_limit()
        if len(s) > limit:
            if self.dom_term:
                s = self.color.py_syntax_highlight(s)
                s = self.dom_term.fold_text_string("", s)
            else:
                s = s[: limit - 3]  # cut before syntax highlighting, to avoid missing color endings
                s = self.color.py_syntax_highlight(s)
                s += "..."
        else:
            s = self.color.py_syntax_highlight(s)
        extra_info = self._pp_extra_info(obj)
        if extra_info != "":
            s += ", " + self.color.py_syntax_highlight(extra_info)
        return s


# For compatibility, we keep non-PEP8 argument names.
# noinspection PyPep8Naming
def format_tb(
    tb=None,
    limit=None,
    allLocals=None,
    allGlobals=None,
    withTitle=False,
    with_color=None,
    with_vars=None,
    clear_frames=True,
):
    """
    :param types.TracebackType|types.FrameType|StackSummary tb: traceback. if None, will use sys._getframe
    :param int|None limit: limit the traceback to this number of frames. by default, will look at sys.tracebacklimit
    :param dict[str,typing.Any]|None allLocals: if set, will update it with all locals from all frames
    :param dict[str,typing.Any]|None allGlobals: if set, will update it with all globals from all frames
    :param bool withTitle:
    :param bool|None with_color: output with ANSI escape codes for color
    :param bool with_vars: will print var content which are referenced in the source code line. by default enabled.
    :param bool clear_frames: whether to call frame.clear() after processing it.
        That will potentially fix some mem leaks regarding locals, so it can be important.
        Also see https://github.com/python/cpython/issues/113939.
        However, any further access to frame locals will not work (e.g. if you want to use a debugger afterwards).
    :return: list of strings (line-based)
    :rtype: list[str]
    """
    color = Color(enable=with_color)
    output = _OutputLinesCollector(color=color)

    def format_filename(s):
        """
        :param str s:
        :rtype: str
        """
        base = os.path.basename(s)
        return (
            color('"' + s[: -len(base)], color.fg_colors[2])
            + color(base, color.fg_colors[2], bold=True)
            + color('"', color.fg_colors[2])
        )

    format_py_obj = output.pretty_print
    if tb is None:
        # noinspection PyBroadException
        try:
            tb = get_current_frame()
            assert tb
        except Exception:
            output(color("format_tb: tb is None and sys._getframe() failed", color.fg_colors[1], bold=True))
            return output.lines

    def is_stack_summary(_tb):
        """
        :param StackSummary|object _tb:
        :rtype: bool
        """
        return isinstance(_tb, StackSummary)

    isframe = inspect.isframe
    if withTitle:
        if isframe(tb) or is_stack_summary(tb):
            output(color("Traceback (most recent call first):", color.fg_colors[0]))
        else:  # expect traceback-object (or compatible)
            output(color("Traceback (most recent call last):", color.fg_colors[0]))
    if with_vars is None and is_at_exit():
        # Better to not show __repr__ of some vars, as this might lead to crashes
        # when native extensions are involved.
        with_vars = False
        if withTitle:
            output("(Exclude vars because we are exiting.)")
    if with_vars is None:
        if any([f.f_code.co_name == "__del__" for f in iter_traceback()]):
            # __del__ is usually called via the Python garbage collector (GC).
            # This can happen and very random / non-deterministic places.
            # There are cases where it is not safe to access some of the vars on the stack
            # because they might be in a non-well-defined state, thus calling their __repr__ is not safe.
            # See e.g. this bug:
            # https://github.com/tensorflow/tensorflow/issues/22770
            with_vars = False
            if withTitle:
                output("(Exclude vars because we are on a GC stack.)")
    if with_vars is None:
        with_vars = True
    # noinspection PyBroadException
    try:
        if limit is None:
            if hasattr(sys, "tracebacklimit"):
                limit = sys.tracebacklimit
        n = 0
        _tb = tb

        class NotFound(Exception):
            """
            Identifier not found.
            """

        def _resolve_identifier(namespace, keys):
            """
            :param dict[str,typing.Any] namespace:
            :param typing.Sequence[str] keys:
            :return: namespace[name[0]][name[1]]...
            """
            if keys[0] not in namespace:
                raise NotFound()
            obj = namespace[keys[0]]
            for part in keys[1:]:
                obj = getattr(obj, part)
            return obj

        # noinspection PyShadowingNames
        def _try_set(old, prefix, func):
            """
            :param None|str old:
            :param str prefix:
            :param func:
            :return: old
            """
            if old is not None:
                return old
            try:
                return add_indent_lines(prefix, func())
            except NotFound:
                return old
            except Exception as e:
                return prefix + "!" + e.__class__.__name__ + ": " + str(e)

        while _tb is not None and (limit is None or n < limit):
            if isframe(_tb):
                f = _tb
            elif is_stack_summary(_tb):
                if isinstance(_tb[0], ExtendedFrameSummary):
                    f = _tb[0].tb_frame
                else:
                    f = DummyFrame.from_frame_summary(_tb[0])
            else:
                f = _tb.tb_frame
            if allLocals is not None:
                allLocals.update(f.f_locals)
            if allGlobals is not None:
                allGlobals.update(f.f_globals)
            if hasattr(_tb, "tb_lineno"):
                lineno = _tb.tb_lineno
            elif is_stack_summary(_tb):
                lineno = _tb[0].lineno
            else:
                lineno = f.f_lineno
            co = f.f_code
            filename = co.co_filename
            if not os.path.isfile(filename):
                alt_fn = fallback_findfile(filename)
                if alt_fn:
                    filename = alt_fn
            name = get_func_str_from_code_object(co, frame=f)
            file_descr = "".join(
                [
                    "  ",
                    color("File ", color.fg_colors[0], bold=True),
                    format_filename(filename),
                    ", ",
                    color("line ", color.fg_colors[0]),
                    color("%d" % lineno, color.fg_colors[4]),
                    ", ",
                    color("in ", color.fg_colors[0]),
                    name,
                ]
            )
            with output.fold_text_ctx(file_descr):
                source_code = get_source_code(filename, lineno, f.f_globals)
                if source_code:
                    source_code = remove_indent_lines(replace_tab_indents(source_code)).rstrip()
                    output("    line: ", color.py_syntax_highlight(source_code), color=color.fg_colors[0])
                    if not with_vars:
                        pass
                    elif isinstance(f, DummyFrame) and not f.have_vars_available:
                        pass
                    else:
                        with output.fold_text_ctx(color("    locals:", color.fg_colors[0])):
                            already_printed_locals = set()  # type: typing.Set[typing.Tuple[str,...]]
                            for token_str in grep_full_py_identifiers(parse_py_statement(source_code)):
                                splitted_token = tuple(token_str.split("."))
                                for token in [splitted_token[0:i] for i in range(1, len(splitted_token) + 1)]:
                                    if token in already_printed_locals:
                                        continue
                                    token_value = None
                                    token_value = _try_set(
                                        token_value,
                                        color("<local> ", color.fg_colors[0]),
                                        lambda: format_py_obj(_resolve_identifier(f.f_locals, token)),
                                    )
                                    token_value = _try_set(
                                        token_value,
                                        color("<global> ", color.fg_colors[0]),
                                        lambda: format_py_obj(_resolve_identifier(f.f_globals, token)),
                                    )
                                    token_value = _try_set(
                                        token_value,
                                        color("<builtin> ", color.fg_colors[0]),
                                        lambda: format_py_obj(_resolve_identifier(f.f_builtins, token)),
                                    )
                                    token_value = token_value or color("<not found>", color.fg_colors[0])
                                    prefix = "      %s " % color(".", color.fg_colors[0], bold=True).join(
                                        token
                                    ) + color("= ", color.fg_colors[0], bold=True)
                                    output(prefix, token_value)
                                    already_printed_locals.add(token)
                            if len(already_printed_locals) == 0:
                                output(color("       no locals", color.fg_colors[0]))
                else:
                    output(color("    -- code not available --", color.fg_colors[0]))

            if clear_frames:
                # Just like :func:`traceback.clear_frames`, but has an additional fix
                # (https://github.com/python/cpython/issues/113939).
                try:
                    f.clear()
                except RuntimeError:
                    pass
                else:
                    # Using this code triggers that the ref actually goes out of scope, otherwise it does not!
                    # https://github.com/python/cpython/issues/113939
                    f.f_locals  # noqa

            if isframe(_tb):
                _tb = _tb.f_back
            elif is_stack_summary(_tb):
                _tb = StackSummary.from_list(_tb[1:])
                if not _tb:
                    _tb = None
            else:
                _tb = _tb.tb_next
            n += 1

    except Exception:
        output(color("ERROR: cannot get more detailed exception info because:", color.fg_colors[1], bold=True))
        import traceback

        for line in traceback.format_exc().split("\n"):
            output("   " + line)

    return output.lines


def print_tb(tb, file=None, **kwargs):
    """
    Replacement for traceback.print_tb.

    :param types.TracebackType|types.FrameType|StackSummary tb:
    :param io.TextIOBase|io.StringIO|typing.TextIO|None file: stderr by default
    :return: nothing, prints to ``file``
    """
    if file is None:
        file = sys.stderr
    for line in format_tb(tb=tb, **kwargs):
        file.write(line)
    file.flush()


def print_exception(etype, value, tb, limit=None, file=None, chain=True):
    """
    Replacement for traceback.print_exception.

    :param etype: exception type
    :param value: exception value
    :param tb: traceback
    :param int|None limit:
    :param io.TextIOBase|io.StringIO|typing.TextIO|None file: stderr by default
    :param bool chain: whether to print the chain of exceptions
    """
    better_exchook(etype, value, tb, autodebugshell=False, file=file, limit=limit, chain=chain)


def print_exc(limit=None, file=None, chain=True):
    """
    Replacement for traceback.print_exc.
    Shorthand for 'print_exception(*sys.exc_info(), limit, file)'.

    :param int|None limit:
    :param io.TextIOBase|io.StringIO|typing.TextIO|None file: stderr by default
    :param bool chain:
    """
    print_exception(*sys.exc_info(), limit=limit, file=file, chain=chain)


def better_exchook(
    etype,
    value,
    tb,
    debugshell=False,
    autodebugshell=True,
    file=None,
    with_color=None,
    with_preamble=True,
    limit=None,
    chain=True,
):
    """
    Replacement for sys.excepthook.

    :param etype: exception type
    :param value: exception value
    :param tb: traceback
    :param bool debugshell: spawn a debug shell at the context of the exception
    :param bool autodebugshell: if env DEBUG is an integer != 0, it will spawn a debug shell
    :param io.TextIOBase|io.StringIO|typing.TextIO|None file: output stream where we will print the traceback
        and exception information. stderr by default.
    :param bool|None with_color: whether to use ANSI escape codes for colored output
    :param bool with_preamble: print a short preamble for the exception
    :param int|None limit:
    :param bool chain: whether to print the chain of exceptions
    """
    if file is None:
        file = sys.stderr

    if autodebugshell:
        # noinspection PyBroadException
        try:
            debugshell = int(os.environ["DEBUG"]) != 0
        except Exception:
            pass

    color = Color(enable=with_color)
    output = _OutputLinesCollector(color=color)

    rec_args = dict(autodebugshell=False, file=file, with_color=with_color, with_preamble=with_preamble)
    if chain:
        if getattr(value, "__cause__", None):
            better_exchook(type(value.__cause__), value.__cause__, value.__cause__.__traceback__, **rec_args)
            output("")
            output("The above exception was the direct cause of the following exception:")
            output("")
        elif getattr(value, "__context__", None):
            better_exchook(type(value.__context__), value.__context__, value.__context__.__traceback__, **rec_args)
            output("")
            output("During handling of the above exception, another exception occurred:")
            output("")

    def format_filename(s):
        """
        :param str s:
        :rtype: str
        """
        base = os.path.basename(s)
        return (
            color('"' + s[: -len(base)], color.fg_colors[2])
            + color(base, color.fg_colors[2], bold=True)
            + color('"', color.fg_colors[2])
        )

    if with_preamble:
        output(color("EXCEPTION", color.fg_colors[1], bold=True))
    all_locals, all_globals = {}, {}
    if tb is not None:
        output.lines.extend(
            format_tb(
                tb=tb,
                limit=limit,
                allLocals=all_locals,
                allGlobals=all_globals,
                withTitle=True,
                with_color=color.enable,
                clear_frames=not debugshell,
            )
        )
    else:
        output(color("better_exchook: traceback unknown", color.fg_colors[1]))

    if isinstance(value, SyntaxError):
        # The standard except hook will also print the source of the SyntaxError,
        # so do it in a similar way here as well.
        filename = value.filename
        # Keep the output somewhat consistent with format_tb.
        file_descr = "".join(
            [
                "  ",
                color("File ", color.fg_colors[0], bold=True),
                format_filename(filename),
                ", ",
                color("line ", color.fg_colors[0]),
                color("%d" % value.lineno, color.fg_colors[4]),
            ]
        )
        with output.fold_text_ctx(file_descr):
            if not os.path.isfile(filename):
                alt_fn = fallback_findfile(filename)
                if alt_fn:
                    output(
                        color("    -- couldn't find file, trying this instead: ", color.fg_colors[0])
                        + format_filename(alt_fn)
                    )
                    filename = alt_fn
            source_code = get_source_code(filename, value.lineno)
            if source_code:
                # Similar to remove_indent_lines.
                # But we need to know the indent-prefix such that we can use the syntax-error offset.
                source_code = replace_tab_indents(source_code)
                lines = source_code.splitlines(True)
                indent_prefix = get_same_indent_prefix(lines)
                if indent_prefix is None:
                    indent_prefix = ""
                source_code = "".join([line[len(indent_prefix) :] for line in lines])
                source_code = source_code.rstrip()
                prefix = "    line: "
                output(prefix, color.py_syntax_highlight(source_code), color=color.fg_colors[0])
                output(" " * (len(prefix) + value.offset - len(indent_prefix) - 1) + "^", color=color.fg_colors[4])

    import types

    # noinspection PyShadowingNames
    def _some_str(value):
        """
        :param object value:
        :rtype: str
        """
        # noinspection PyBroadException
        try:
            return str(value)
        except Exception:
            return "<unprintable %s object>" % type(value).__name__

    # noinspection PyShadowingNames
    def _format_final_exc_line(etype, value):
        value_str = _some_str(value)
        if value is None or not value_str:
            line = color("%s" % etype, color.fg_colors[1])
        else:
            line = color("%s" % etype, color.fg_colors[1]) + ": %s" % (value_str,)
        return line

    # noinspection PyUnresolvedReferences
    if (
        isinstance(etype, BaseException)
        or (hasattr(types, "InstanceType") and isinstance(etype, types.InstanceType))
        or etype is None
        or type(etype) is str
    ):
        output(_format_final_exc_line(etype, value))
    else:
        output(_format_final_exc_line(etype.__name__, value))

    for line in output.lines:
        file.write(line)
    file.flush()

    if debugshell:
        file.write("---------- DEBUG SHELL -----------\n")
        file.flush()
        debug_shell(user_ns=all_locals, user_global_ns=all_globals, traceback=tb)


def dump_all_thread_tracebacks(exclude_thread_ids=None, file=None):
    """
    Prints the traceback of all threads.

    :param set[int]|list[int]|None exclude_thread_ids: threads to exclude
    :param io.TextIOBase|io.StringIO|typing.TextIO|None file: output stream
    """
    if exclude_thread_ids is None:
        exclude_thread_ids = []
    if not file:
        file = sys.stdout
    import threading

    if hasattr(sys, "_current_frames"):
        print("", file=file)
        threads = {t.ident: t for t in threading.enumerate()}
        # noinspection PyProtectedMember
        for tid, stack in sys._current_frames().items():
            if tid in exclude_thread_ids:
                continue
            # This is a bug in earlier Python versions.
            # https://bugs.python.org/issue17094
            # Note that this leaves out all threads not created via the threading module.
            if tid not in threads:
                continue
            tags = []
            thread = threads.get(tid)
            if thread:
                assert isinstance(thread, threading.Thread)
                if thread is threading.current_thread():
                    tags += ["current"]
                # noinspection PyProtectedMember,PyUnresolvedReferences
                if isinstance(thread, threading._MainThread):
                    tags += ["main"]
                tags += [str(thread)]
            else:
                tags += ["unknown with id %i" % tid]
            print("Thread %s:" % ", ".join(tags), file=file)
            print_tb(stack, file=file)
            print("", file=file)
        print("That were all threads.", file=file)
    else:
        print("Does not have sys._current_frames, cannot get thread tracebacks.", file=file)


def get_current_frame():
    """
    :return: current frame object (excluding this function call)
    :rtype: types.FrameType

    Uses sys._getframe if available, otherwise some trickery with sys.exc_info and a dummy exception.
    """
    if hasattr(sys, "_getframe"):
        # noinspection PyProtectedMember
        return sys._getframe(1)
    try:
        raise ZeroDivisionError
    except ZeroDivisionError:
        return sys.exc_info()[2].tb_frame.f_back


def get_func_str_from_code_object(co, frame=None):
    """
    :param types.CodeType co:
    :param types.FrameType|DummyFrame|None frame: if given, might provide a faster way to get the function name
    :return: co.co_name as fallback, but maybe sth better like the full func name if possible
    :rtype: str
    """
    f = get_func_from_code_object(co, frame=frame)
    if f:
        if hasattr(f, "__qualname__"):
            return f.__qualname__
        return str(f)
    return co.co_name


def get_func_from_code_object(co, frame=None):
    """
    :param types.CodeType co:
    :param types.FrameType|DummyFrame|None frame: if given, might provide a faster way to get the function name
    :return: function, such that ``func.__code__ is co``, or None
    :rtype: types.FunctionType

    This is CPython specific (to some degree; it uses the `gc` module to find references).
    Inspired from:
    https://stackoverflow.com/questions/12787108/getting-the-python-function-for-a-code-object
    https://stackoverflow.com/questions/54656758/get-function-object-from-stack-frame-object
    """
    import gc
    import types

    assert isinstance(co, (types.CodeType, DummyFrame))
    _attr_name = "__code__" if PY3 else "func_code"
    if frame and frame.f_code.co_nlocals > 0:
        func_name = frame.f_code.co_name
        frame_self = frame.f_locals.get("self")
        if frame_self is not None:
            candidate = getattr(frame_self.__class__, func_name, None)
            if candidate and (getattr(candidate, _attr_name, None) is co or isinstance(co, DummyFrame)):
                return candidate
    try:
        candidate = getattr(_get_loaded_module_from_filename(co.co_filename), co.co_name, None)
    except ImportError:  # some modules have lazy loaders, but those might fail here
        candidate = None
    if candidate and (getattr(candidate, _attr_name, None) is co or isinstance(co, DummyFrame)):
        return candidate
    if isinstance(co, DummyFrame):
        return None
    candidates = gc.get_referrers(co)
    candidates = [f for f in candidates if getattr(f, _attr_name, None) is co]
    if candidates:
        return candidates[0]
    return None


_loaded_module_from_filename_cache = {}  # filename -> module name


def _get_loaded_module_from_filename(filename):
    """
    Like inspect.getmodule but faster.

    :param str filename:
    :rtype: types.ModuleType|Any|None
    """
    if filename.endswith(".pyc") or filename.endswith(".pyo"):
        filename = filename[:-1]
    if filename in _loaded_module_from_filename_cache:
        return sys.modules.get(_loaded_module_from_filename_cache[filename])
    # Update the filename to module name cache and check yet again
    # Copy sys.modules in order to cope with changes while iterating
    for modname, module in sys.modules.copy().items():
        f = getattr(module, "__file__", None)
        if f:
            if f.endswith(".pyc") or f.endswith(".pyo"):
                f = f[:-1]
            _loaded_module_from_filename_cache[f] = modname
    if filename in _loaded_module_from_filename_cache:
        return sys.modules.get(_loaded_module_from_filename_cache[filename])
    return None


def iter_traceback(tb=None, enforce_most_recent_call_first=False):
    """
    Iterates a traceback of various formats:
      - traceback (types.TracebackType)
      - frame object (types.FrameType)
      - stack summary (traceback.StackSummary)

    :param types.TracebackType|types.FrameType|StackSummary|None tb: traceback. if None, will use sys._getframe
    :param bool enforce_most_recent_call_first:
        Frame or stack summery: most recent call first (top of the stack is the first entry in the result)
        Traceback: most recent call last
        If True, and we get traceback, will unroll and reverse, such that we have always the most recent call first.
    :return: yields the frames (types.FrameType)
    :rtype: list[types.FrameType|DummyFrame]
    """
    if tb is None:
        tb = get_current_frame()

    def is_stack_summary(_tb):
        """
        :param StackSummary|object _tb:
        :rtype: bool
        """
        return isinstance(_tb, StackSummary)

    is_frame = inspect.isframe
    is_traceback = inspect.istraceback
    assert is_traceback(tb) or is_frame(tb) or is_stack_summary(tb)
    # Frame or stack summery: most recent call first
    # Traceback: most recent call last
    if is_traceback(tb) and enforce_most_recent_call_first:
        frames = list(iter_traceback(tb))
        for frame in frames[::-1]:
            yield frame
        return

    _tb = tb
    while _tb is not None:
        if is_frame(_tb):
            frame = _tb
        elif is_stack_summary(_tb):
            if isinstance(_tb[0], ExtendedFrameSummary):
                frame = _tb[0].tb_frame
            else:
                frame = DummyFrame.from_frame_summary(_tb[0])
        else:
            frame = _tb.tb_frame
        yield frame
        if is_frame(_tb):
            _tb = _tb.f_back
        elif is_stack_summary(_tb):
            _tb = StackSummary.from_list(_tb[1:])
            if not _tb:
                _tb = None
        else:
            _tb = _tb.tb_next


class ExtendedFrameSummary(FrameSummary):
    """
    Extends :class:`FrameSummary` by ``self.tb_frame``.
    """

    def __init__(self, frame, **kwargs):
        super(ExtendedFrameSummary, self).__init__(**kwargs)
        self.tb_frame = frame


class DummyFrame:
    """
    This class has the same attributes as a code and a frame object
    and is intended to be used as a dummy replacement.
    """

    @classmethod
    def from_frame_summary(cls, f):
        """
        :param FrameSummary f:
        :rtype: DummyFrame
        """
        return cls(filename=f.filename, lineno=f.lineno, name=f.name, f_locals=f.locals)

    def __init__(self, filename, lineno, name, f_locals=None, f_globals=None, f_builtins=None):
        self.lineno = lineno
        self.tb_lineno = lineno
        self.f_lineno = lineno
        self.f_code = self
        self.filename = filename
        self.co_filename = filename
        self.name = name
        self.co_name = name
        self.f_locals = f_locals or {}
        self.f_globals = f_globals or {}
        self.f_builtins = f_builtins or {}
        self.have_vars_available = f_locals is not None or f_globals is not None or f_builtins is not None
        self.co_nlocals = len(self.f_locals) if self.f_locals is not self.f_globals else 0

    def clear(self):
        """clear"""
        self.f_locals = None


# noinspection PyPep8Naming,PyUnusedLocal
def _StackSummary_extract(frame_gen, limit=None, lookup_lines=True, capture_locals=False):
    """
    Replacement for :func:`StackSummary.extract`.

    Create a StackSummary from a traceback or stack object.
    Very simplified copy of the original StackSummary.extract().
    We want always to capture locals, that is why we overwrite it.
    Additionally, we also capture the frame.
    This is a bit hacky and also not like this is originally intended (to not keep refs).

    :param frame_gen: A generator that yields (frame, lineno) tuples to
        include in the stack.
    :param limit: None to include all frames or the number of frames to
        include.
    :param lookup_lines: If True, lookup lines for each frame immediately,
        otherwise lookup is deferred until the frame is rendered.
    :param capture_locals: If True, the local variables from each frame will
        be captured as object representations into the FrameSummary.
    """
    result = StackSummary()
    for f, lineno in frame_gen:
        co = f.f_code
        filename = co.co_filename
        name = co.co_name
        result.append(ExtendedFrameSummary(frame=f, filename=filename, lineno=lineno, name=name, lookup_line=False))
    return result


def install():
    """
    Replaces sys.excepthook by our better_exchook.
    """
    sys.excepthook = better_exchook


def replace_traceback_format_tb():
    """
    Replaces these functions from the traceback module by our own:

    - traceback.format_tb
    - traceback.StackSummary.format
    - traceback.StackSummary.extract

    Note that this kind of monkey patching might not be safe under all circumstances
    and is not officially supported by Python.
    """
    import traceback

    traceback.format_tb = format_tb
    if hasattr(traceback, "StackSummary"):
        traceback.StackSummary.format = format_tb
        traceback.StackSummary.extract = _StackSummary_extract


def replace_traceback_print_tb():
    """
    Replaces these functions from the traceback module by our own:

    - traceback.print_tb
    - traceback.print_exception
    - traceback.print_exc

    Note that this kind of monkey patching might not be safe under all circumstances
    and is not officially supported by Python.
    """
    import traceback

    traceback.print_tb = print_tb
    traceback.print_exception = print_exception
    traceback.print_exc = print_exc


def setup_all():
    """
    Calls:

    - :func:`install`
    - :func:`replace_traceback_format_tb`
    - :func:`replace_traceback_print_tb`
    """
    install()
    replace_traceback_format_tb()
    replace_traceback_print_tb()
